import logging

from django.conf import settings
from django.dispatch import Signal
from django.utils.module_loading import import_string

from terracommon.events.models import EventHandler

logger = logging.getLogger(__name__)

event = Signal(providing_args=['action', 'logged_user'])


def signal_event_proxy(sender, action, instance, user, *args, **kwargs):
    for handler in (EventHandler.objects
                                .filter(action=action)
                                .order_by('priority')):
        try:
            args = {
                'instance': instance,
                'user': user,
                **kwargs
            }

            handler_class = import_string(handler.handler)
            try:
                executor = handler_class(handler.action,
                                         handler.settings,
                                         **args)

                if executor.valid_condition():
                    executor()
            except Exception as e:
                if settings.DEBUG:
                    raise
                else:
                    logger.error('Handler error: %s',
                                 e,
                                 extra={'handler': handler.handler})
        except ImportError as e:
            logger.error(f"An error occured loading {handler.handler}: {e}")


event.connect(signal_event_proxy)



# Generated by Django 2.0.7 on 2018-09-11 08:27

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('document_generator', '0002_documenttemplate_uid'),
    ]

    operations = [
        migrations.AlterField(
            model_name='documenttemplate',
            name='uid',
            field=models.CharField(default=uuid.uuid4, max_length=256),
        ),
    ]



# Generated by Django 2.0.9 on 2018-11-20 10:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('trrequests', '0002_auto_20180824_1017'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='comment',
            options={'ordering': ['id']},
        ),
        migrations.AlterModelOptions(
            name='userrequest',
            options={'ordering': ['id'], 'permissions': (('can_create_requests', 'Is able to create a new requests'), ('can_read_self_requests', 'Is able to get own requests'), ('can_read_all_requests', 'Is able to get all requests'), ('can_comment_requests', 'Is able to comment an user request'), ('can_internal_comment_requests', 'Is able to add comments not visible by users'), ('can_read_comment_requests', 'Is allowed to read only non-internal comments'), ('can_change_state_requests', 'Is authorized to change the request state'), ('can_download_all_pdf', 'Is able to download a pdf document'))},
        ),
    ]



# Generated by Django 2.0.13 on 2019-07-11 16:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tropp', '0003_auto_20190122_1551'),
    ]

    operations = [
        migrations.AlterField(
            model_name='picture',
            name='date',
            field=models.DateTimeField(null=True, verbose_name='Date'),
        ),
    ]



from pprint import pformat

import coreapi
import coreschema
from django.conf import settings
from django.contrib.postgres.fields.jsonb import KeyTextTransform
from django.db.models import Q
from rest_framework import filters
from rest_framework.exceptions import ValidationError

from terracommon.core.settings import STATES


class CampaignFilterBackend(filters.BaseFilterBackend):
    """
    Filters for campaigns
    """

    def get_schema_fields(self, view):
        super().get_schema_fields(view)
        choices = {
            STATES.DRAFT: 'Incomplete metadata',
            STATES.SUBMITTED: 'Pending validation',
            STATES.REFUSED: 'Refused',
            STATES.ACCEPTED: 'Validated',
        }
        return [
            coreapi.Field(
                name='status',
                required=False,
                location='query',
                schema=coreschema.Boolean(
                    title="Campaign status",
                    description="0 for closed campaign, 1 for ongoing campaign"
                ),
            ),
            coreapi.Field(
                name='picture_status',
                required=False,
                location='query',
                schema=coreschema.Enum(
                    choices,
                    description=str(pformat(choices)),
                    title="Picture status",
                )
            )
        ]

    def filter_queryset(self, request, queryset, view):
        status = request.GET.get('status', None)
        if status is not None:
            try:
                status = bool(int(status))
            except ValueError:
                raise ValidationError
            if status:
                queryset = queryset.filter(
                    viewpoints__pictures__state=STATES.ACCEPTED
                )
            else:
                queryset = queryset.exclude(
                    Q(viewpoints__pictures__state=STATES.ACCEPTED)
                )

        picture_status = request.GET.get('picture_status', None)
        if picture_status is not None:
            try:
                picture_status = int(picture_status)
                assert picture_status in STATES.CHOICES_DICT
            except (AssertionError, ValueError):
                raise ValidationError
            queryset = queryset.filter(
                viewpoints__pictures__state=picture_status
            )

        return queryset


class JsonFilterBackend(filters.BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        for key, field in settings.TROPP_SEARCHABLE_PROPERTIES.items():
            search_key = f'properties__{field["json_key"]}'
            if field['type'] == 'many':
                search_item = request.GET.getlist(f'{search_key}[]')
            else:
                search_item = request.GET.get(search_key)
            if search_item:
                if field['type'] == 'text':
                    # Full text search
                    queryset = queryset.annotate(**{
                        key: KeyTextTransform(field['json_key'], 'properties')
                    }).filter(**{f'{key}__icontains': search_item})
                else:
                    # Search on element
                    queryset = queryset.filter(**{
                        f'{search_key}__contains': search_item
                    })
        return queryset

    def get_schema_fields(self, view):
        fields = []
        for key, field in settings.TROPP_SEARCHABLE_PROPERTIES.items():
            if field['type'] == 'many':
                klass = coreschema.Array
            else:
                klass = coreschema.String
            description = f"{key.capitalize()} property ({field['type']})"
            fields.append(
                coreapi.Field(
                    name=f'properties__{field["json_key"]}',
                    required=False,
                    location='query',
                    schema=klass(
                        title=f"{key.capitalize()} property",
                        description=description,
                    ),
                )
            )
        return super().get_schema_fields(view)



# Generated by Django 2.0.13 on 2019-04-17 16:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('terra', '0021_auto_20190415_0945'),
    ]

    operations = [
        migrations.AlterField(
            model_name='feature',
            name='source',
            field=models.IntegerField(blank=True, help_text='Internal field used by pgRouting', null=True),
        ),
        migrations.AlterField(
            model_name='feature',
            name='target',
            field=models.IntegerField(blank=True, help_text='Internal field used by pgRouting', null=True),
        ),
    ]




from django.contrib.auth.models import Group, Permission
from django.test import TestCase
from django.urls import reverse
from rest_framework.status import (HTTP_200_OK, HTTP_401_UNAUTHORIZED,
                                   HTTP_403_FORBIDDEN)
from rest_framework.test import APIClient

from terracommon.accounts.tests.factories import TerraUserFactory

from .models import DataStore, DataStorePermission


class DataStoreTestCase(TestCase):
    def setUp(self):

        examples = [{
            'key': 'terracommon.test.data.store',
            'value': {'key': 'value', }
        }, {
            'key': 'terracommon.test.data.new_store',
            'value': {}
        }, {
            'key': 'terracommon.prefix.dot.com',
            'value': {}
        }]

        for example in examples:
            DataStore.objects.create(**example)

        self.client = APIClient()
        self.user = TerraUserFactory()
        self.client.force_authenticate(user=self.user)

    def test_not_authenticated(self):
        client = APIClient()
        response = client.get(reverse('datastore:datastore-list'))
        self.assertEqual(HTTP_401_UNAUTHORIZED, response.status_code)

    def test_no_permission(self):
        client = APIClient()
        client.force_authenticate(user=self.user)
        response = client.get(reverse('datastore:datastore-list'))

        self.assertEqual(HTTP_200_OK, response.status_code)
        self.assertEqual(0, response.json()['count'])

    def test_can_readonly(self):
        # Add the user to a group
        group = Group.objects.create(name='can_read')
        self.user.groups.add(group)
        self.user.save()

        # Allow this group to read the terracommon.prefix prefix
        perm = Permission.objects.get(codename='can_read_datastore')
        DataStorePermission.objects.create(
            group=group,
            permission=perm,
            prefix='terracommon.prefix',
        )

        response = self.client.get(reverse('datastore:datastore-list'))
        self.assertEqual(HTTP_200_OK, response.status_code)
        self.assertEqual(1, response.json()['count'])

        # test that write is not allowed
        test_value = {'a': 'TEST'}
        ds = DataStore.objects.get(key='terracommon.prefix.dot.com')
        response = self.client.put(reverse('datastore:datastore-detail',
                                           args=[ds.key]),
                                   data={'value': test_value})

        self.assertEqual(HTTP_403_FORBIDDEN, response.status_code)
        self.user.groups.clear()

    def test_can_readwrite(self):
        # Add the user to a group
        group = Group.objects.create(name='can_readwrite')
        self.user.groups.add(group)

        # Allow this group to read the terracommon.prefix prefix
        perm = Permission.objects.get(codename='can_readwrite_datastore')
        DataStorePermission.objects.get_or_create(
            group=group,
            permission=perm,
            prefix='terracommon.prefix',
        )

        response = self.client.get(reverse('datastore:datastore-list'))
        self.assertEqual(HTTP_200_OK, response.status_code)
        self.assertEqual(1, response.json()['count'])

        # Test writing
        test_value = {'a': 'b'}
        ds = DataStore.objects.get(key='terracommon.prefix.dot.com')
        response = self.client.put(reverse('datastore:datastore-detail',
                                           args=[ds.key]),
                                   data=test_value)

        self.assertEqual(HTTP_200_OK, response.status_code)

        ds.refresh_from_db()
        self.assertDictEqual(ds.value, test_value)

        self.user.groups.clear()

    def test_create_item(self):
        # Add the user to a group
        group = Group.objects.create(name='can_readwrite')
        self.user.groups.add(group)

        # Allow this group to read the terracommon.prefix prefix
        perm = Permission.objects.get(codename='can_readwrite_datastore')
        DataStorePermission.objects.get_or_create(
            group=group,
            permission=perm,
            prefix='terracommon.prefix',
        )

        test_value = {'test_key': 'test value'}
        response = self.client.post(reverse('datastore:datastore-detail', args=['terracommon.prefix.blurp']),
                                    data=test_value)
        self.assertEqual(response.status_code, HTTP_200_OK)

        self.assertTrue(DataStore.objects.filter(key='terracommon.prefix.blurp').exists())
        self.assertDictEqual(response.json(), test_value)

    def test_forbidden_create_item(self):
        user = TerraUserFactory()
        client = APIClient()
        client.force_authenticate(user=user)

        response = client.post(reverse('datastore:datastore-detail',
                                       args=['terracommon.prefix.forbidden_creation']),
                               data={'test_key': 'test value'})
        self.assertEqual(response.status_code, HTTP_403_FORBIDDEN)
        self.assertFalse(DataStore.objects.filter(key='terracommon.prefix.forbidden_creation').exists())

    def test_perms(self):
        # Add the user to a group
        group = Group.objects.create(name='can_readwrite')
        self.user.groups.add(group)

        # Allow this group to read the terracommon.prefix prefix
        perm = Permission.objects.get(codename='can_readwrite_datastore')
        DataStorePermission.objects.get_or_create(
            group=group,
            permission=perm,
            prefix='terracommon.testprefix',
        )

        test_value = {'test_key': 'test value'}
        response = self.client.post(reverse('datastore:datastore-detail', args=['terracommon.testprefix.blurp']),
                                    data=test_value)
        self.assertEqual(response.status_code, HTTP_200_OK)

        response = self.client.post(reverse('datastore:datastore-detail', args=['terracommon.forbiddenprefix']),
                                    data=test_value)
        self.assertEqual(response.status_code, HTTP_403_FORBIDDEN)



from django.urls import path
from rest_framework.routers import SimpleRouter

from . import views

app_name = 'tropp'

router = SimpleRouter()

router.register(r'viewpoints', views.ViewpointViewSet, base_name='viewpoint')
router.register(r'campaigns', views.CampaignViewSet, base_name='campaign')
router.register(r'pictures', views.PictureViewSet, base_name='picture')

urlpatterns = router.urls

urlpatterns += [
    path(
        'viewpoints/<int:pk>/pdf',
        views.ViewpointPdf.as_view(),
        name='viewpoint-pdf',
    ),
    path(
        'viewpoints/<int:pk>/zip-pictures',
        views.ViewpointZipPictures.as_view(),
        name='viewpoint-zip',
    ),
]



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.utils.translation import ugettext_lazy as _

from terracommon.accounts.models import TerraUser


class TerraUserAdmin(UserAdmin):
    ordering = ('email',)
    list_display = ('email', 'is_staff')
    search_fields = ('email', )
    fieldsets = (
        (None, {'fields': ('email', 'password')}),
        (_('Permissions'), {'fields': ('is_active', 'is_staff', 'is_superuser',
                                       'groups', 'user_permissions')}),
        (_('Important dates'), {'fields': ('last_login', 'date_joined')}),
    )
    add_fieldsets = (
        (None, {
            'classes': ('wide',),
            'fields': ('email', 'password1', 'password2'),
        }),
    )


admin.site.register(TerraUser, TerraUserAdmin)



import logging

from django.contrib.auth import get_user_model
from django.core.management import call_command
from django.test import TestCase

from terracommon.terra.models import Layer
from terracommon.terra.tests.utils import get_files_tests
from terracommon.terra.transformations import set_geometry_from_options

UserModel = get_user_model()


class ImportCSVTestCase(TestCase):
    def test_command_launch(self):
        test_file = get_files_tests('test.csv')
        with self.assertLogs(level=logging.WARNING) as cm:
            call_command('import_csv',
                         ('--operation=terracommon.data_importers.tests'
                          '.test_dataimporter_functions.empty_operation'),
                         ('--operation=terracommon.terra.transformations'
                          '.set_geometry_from_options'),
                         '--layer=companies',
                         '--key=SIREN',
                         '--key=NIC',
                         f'--source={test_file}')
        self.assertEqual(len(cm.records), 1)
        log_record = cm.records[0]
        self.assertEqual(set_geometry_from_options.__name__,
                         log_record.funcName)
        self.assertIn('019778745', log_record.msg)  # SIREN key
        self.assertIn('00018', log_record.msg)  # NIC key
        self.assertIsNotNone(Layer.objects.filter(name="companies"))



from django.contrib.auth.tokens import default_token_generator
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode

from terracommon.core.mixins import SerializerCurrentUserMixin

from .models import ReadModel


class ReadableModelMixin(object):

    def user_read(self, user):
        if not user:
            raise AttributeError('Cannot set element read for unknown user')
        ReadModel.objects.read_object(user, self)

    def get_user_read(self, user):
        return ReadModel.objects.get_user_read(user, self)


class UserTokenGeneratorMixin(SerializerCurrentUserMixin):
    def get_uidb64_token_for_user(self, user):
        if user:
            return (urlsafe_base64_encode(force_bytes(user.pk)).decode(),
                    default_token_generator.make_token(self.current_user))
        else:
            return (None, None)



# Generated by Django 2.0.9 on 2019-01-02 14:08

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('events', '0003_auto_20181120_1059'),
    ]

    operations = [
        migrations.AlterField(
            model_name='eventhandler',
            name='settings',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]



# Generated by Django 2.0.7 on 2018-08-24 10:17

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('documenttemplate', models.FileField(upload_to='templates/%Y/%m')),
            ],
        ),
        migrations.CreateModel(
            name='DownloadableDocument',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='contenttypes.ContentType')),
                ('document', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='document_generator.DocumentTemplate')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# Generated by Django 2.0.7 on 2018-09-04 07:24

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('document_generator', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='documenttemplate',
            name='uid',
            field=models.CharField(default=uuid.UUID('c32ff5a4-3df1-4d1e-9319-7f47c64f4f35'), max_length=256),
        ),
    ]



# Generated by Django 2.0.7 on 2018-07-11 08:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('terra', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='feature',
            name='source',
            field=models.IntegerField(help_text='Internal field used by pgRouting', null=True),
        ),
        migrations.AddField(
            model_name='feature',
            name='target',
            field=models.IntegerField(help_text='Internal field used by pgRouting', null=True),
        ),
    ]



from rest_framework import serializers

from .models import UserNotifications


class UserNotificationSerializer(serializers.ModelSerializer):

    class Meta:
        model = UserNotifications
        exclude = ('user', )



import json

from django.core.management import call_command
from django.db import connection
from django.test import TestCase, override_settings
from django.urls import reverse
from rest_framework.status import (HTTP_200_OK, HTTP_204_NO_CONTENT,
                                   HTTP_404_NOT_FOUND)

from terracommon.terra.models import Layer
from terracommon.terra.tests.factories import LayerFactory
from terracommon.terra.tests.utils import get_files_tests
from terracommon.terra.tiles.helpers import (VectorTile, guess_maxzoom,
                                             guess_minzoom)


@override_settings(DEBUG=True, CACHES={
    'default': {
        'BACKEND': ('django.core.cache.backends'
                    '.locmem.LocMemCache')
    }})
class VectorTilesNoLayerTestCase(TestCase):
    group_name = 'mygroup'

    @override_settings(ALLOWED_HOSTS=['localhost'])
    def test_tilejson_fail_no_layer(self):
        response = self.client.get(
            reverse('terra:group-tilejson', args=[self.group_name]),
            HTTP_HOST='localhost')
        self.assertEqual(HTTP_404_NOT_FOUND, response.status_code)

    def test_vector_tiles_view_without_layer(self):
        # first query that generate the cache
        response = self.client.get(
            reverse('terra:group-tiles', args=[self.group_name, 10, 515, 373]))
        self.assertEqual(HTTP_404_NOT_FOUND, response.status_code)
        self.assertEqual(len(response.content), 0)


@override_settings(DEBUG=True, CACHES={
    'default': {
        'BACKEND': ('django.core.cache.backends'
                    '.locmem.LocMemCache')
    }})
class VectorTilesTestCase(TestCase):
    group_name = 'mygroup'

    def setUp(self):
        settings = {'metadata': {'attribution': 'plop'}}

        self.layer = LayerFactory(group=self.group_name, name="layerLine", settings=settings)

        self.layer.from_geojson(
            geojson_data='''
            {
            "type": "FeatureCollection",
            "features": [
                {
                "type": "Feature",
                "properties": {
                    "foo": "bar",
                    "baba": "fifi"
                },
                "geometry": {
                    "type": "LineString",
                    "coordinates": [
                    [
                        1.3700294494628906,
                        43.603640347220924
                    ],
                    [
                        1.2984466552734375,
                        43.57902295875415
                    ]
                    ]
                }
                }
            ]
            }
        ''')

        self.layerPoint = LayerFactory(group="yourgroup", name="layerPoint", settings=settings)

        self.layerPoint.from_geojson(

            geojson_data='''
            {
            "type": "FeatureCollection",
            "features": [
                {
                "type": "Feature",
                "properties": {
                    "foo": "bar"
                },
                "geometry": {
                    "type": "Point",
                    "coordinates": [
                        1.3700294494628906,
                        43.603640347220924
                    ]
                }
                }
            ]
            }
        ''')

    @override_settings(ALLOWED_HOSTS=['localhost'])
    def test_tilejson(self):
        response = self.client.get(
            reverse('terra:group-tilejson', args=[self.group_name]),
            # HTTP_HOST required to build the tilejson descriptor
            HTTP_HOST='localhost')
        self.assertEqual(HTTP_200_OK, response.status_code)
        self.assertGreater(len(response.content), 0)

        tilejson = json.loads(response.content)
        self.assertTrue(tilejson['attribution'])
        self.assertTrue(tilejson['description'] is None)
        self.assertGreater(len(tilejson['vector_layers']), 0)
        self.assertGreater(len(tilejson['vector_layers'][0]['fields']), 0)

    def test_vector_tiles_view(self):
        # first query that generate the cache
        response = self.client.get(
            reverse('terra:group-tiles', args=[self.group_name, 10, 515, 373]))
        self.assertEqual(HTTP_200_OK, response.status_code)
        self.assertGreater(len(response.content), 0)
        query_count = len(connection.queries)
        original_content = response.content

        # verify data is cached
        response = self.client.get(
            reverse('terra:group-tiles', args=[self.group_name, 10, 515, 373]))
        self.assertEqual(
            len(connection.queries),
            query_count - 2
        )
        self.assertEqual(
            original_content,
            response.content
        )

        response = self.client.get(
            reverse('terra:group-tiles', args=[self.group_name, 10, 1, 1]))

        self.assertEqual(HTTP_200_OK, response.status_code)
        self.assertEqual(b'', response.content)

    @override_settings(MAX_TILE_ZOOM=9)
    def test_vector_tiles_view_max_tile_zoom_lower_actual_zoom(self):
        # first query that generate the cache
        response = self.client.get(
            reverse('terra:group-tiles', args=[self.group_name, 10, 515, 373]))
        self.assertEqual(HTTP_204_NO_CONTENT, response.status_code)
        self.assertEqual(len(response.content), 0)

    def test_filtering(self):
        features = self.layer.features.all()
        tile = VectorTile(self.layer, "CACHINGCACHE")
        x, y, z = 16506, 11966, 15
        pixel_buffer, features_filter, properties_filter, features_limit = \
            4, {'baba': 'fifi2'}, ['foo'], 10000

        tile = tile.get_tile(
            x, y, z,
            pixel_buffer, features_filter, properties_filter, features_limit,
            features)
        self.assertGreater(len(tile), 0)

    def test_guess_maxzoom(self):

        # guess_maxzoom returning -1 when TypeError is raised14)
        self.assertEqual(
            guess_maxzoom(self.layerPoint),
            14)

        self.assertEqual(
            guess_maxzoom(self.layer) is not None,
            True)

        # test guess_maxzoom returns sensible value from OSM Fontainebleau paths&tracks
        chunk_fontainebleau_geojson = get_files_tests('chunk_fontainebleau.geojson')

        call_command(
            'import_geojson', chunk_fontainebleau_geojson,
            '-gr', 'maxzoom_test',
            '-ln', 'chunk_fontainebleau',
            verbosity=0)

        layer_chunk_fontainebleau = Layer.objects.get(name='chunk_fontainebleau')

        self.assertEqual(guess_maxzoom(layer_chunk_fontainebleau), 13)

    def test_guess_minzoom(self):

        self.assertEqual(
            guess_minzoom(self.layerPoint),
            0)

        self.assertTrue(isinstance(
            guess_minzoom(self.layer),
            int)
        )


class VectorTilesSpecialTestCase(TestCase):
    group_name = 'mygroup'

    def setUp(self):
        # Same as default with properties filter not None
        settings = {'metadata': {'attribution': 'plop'}, 'tiles': {
            'minzoom': 0,
            'maxzoom': 22,
            'pixel_buffer': 4,
            'features_filter': None,  # Json
            'properties_filter': 'Test',  # Array of string
            'features_limit': 10000,
        }}

        self.layer = LayerFactory(group=self.group_name, name="layerLine", settings=settings)

        self.geojson_data = '''
            {
            "type": "FeatureCollection",
            "features": [
                {
                "type": "Feature",
                "properties": {
                    "foo": "bar",
                    "baba": "fifi"
                },
                "geometry": {
                    "type": "LineString",
                    "coordinates": [
                    [
                        1.3700294494628906,
                        43.603640347220924
                    ],
                    [
                        1.2984466552734375,
                        43.57902295875415
                    ]
                    ]
                }
                }
            ]
            }
        '''
        self.layer.from_geojson(geojson_data=self.geojson_data)

    @override_settings(ALLOWED_HOSTS=['localhost'])
    def test_tilejson_with_properties(self):
        response = self.client.get(
            reverse('terra:group-tilejson', args=[self.group_name]),
            # HTTP_HOST required to build the tilejson descriptor
            HTTP_HOST='localhost')
        self.assertEqual(HTTP_200_OK, response.status_code)
        self.assertGreater(len(response.content), 0)

        tilejson = json.loads(response.content)
        self.assertTrue(tilejson['attribution'])
        self.assertTrue(tilejson['description'] is None)
        self.assertGreater(len(tilejson['vector_layers']), 0)
        self.assertGreater(len(tilejson['vector_layers'][0]['fields']), 0)



from django.test import TestCase
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APIClient

from terracommon.terra.tests.factories import LayerFactory, UserFactory


class SchemaValidationTest(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserFactory()
        self.client.force_authenticate(user=self.user)

        self.no_schema_layer = LayerFactory(name="no schema", geom_type=None)
        self.valid_schema = {
            "properties": {
                "name": {
                    "type": "string"
                },
                "age": {
                    "type": "integer"
                }
            }
        }
        self.property_schema_layer = LayerFactory(
            name="tree",
            schema=self.valid_schema)

    def test_create_layer_without_valid_schema(self):
        """
        Try to create layer with valid schema
        """
        response = self.client.post(reverse('terra:layer-list'),
                                    data={})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_layer_with_valid_schema(self):
        """
        Try to create layer with valid schema
        """
        response = self.client.post(reverse('terra:layer-list'),
                                    data={"schema": self.valid_schema})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_layer_unvalid_schema(self):
        """
        Try to create layer with unvalid schema
        """
        response = self.client.post(reverse('terra:layer-list'),
                                    data={"schema": {"type": "unknown"}})
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("schema", response_json)

    def test_no_schema_properties_ok(self):
        """
        If no schema defined (or empty), all properties are accepted
        """
        response = self.client.post(reverse('terra:feature-list', args=[self.no_schema_layer.pk, ]),
                                    data={"geom": "POINT(0 0)",
                                          "properties": {"toto": "ok"}})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())

    def test_schema_property_match_good(self):
        """
        If schema defined, allow features with properties in schema
        """
        response = self.client.post(reverse('terra:feature-list', args=[self.property_schema_layer.pk, ]),
                                    data={"geom": "POINT(0 0)",
                                          "properties": {"name": "ok",
                                                         "age": 10}})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())

    def test_schema_property_match_bad(self):
        """
        If schema defined, deny unvalid data
        """
        response = self.client.post(reverse('terra:feature-list', args=[self.property_schema_layer.pk, ]),
                                    data={"geom": "POINT(0 0)",
                                          "properties": {"name": 20,
                                                         "age": "wrong data"}})
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("properties", response_json)
        self.assertIn("wrong data", response_json['properties'][0])

    def test_schema_property_doesnt_match(self):
        """
        If schema defined, deny properties not in schema
        """
        response = self.client.post(reverse('terra:feature-list', args=[self.property_schema_layer.pk, ]),
                                    data={"geom": "POINT(0 0)",
                                          "properties": {"toto": "ok"}})
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("properties", response_json)
        self.assertIn("toto", response_json['properties'][0])



import io
import logging
import os
import zipfile

import magic
from django.conf import settings
from django.contrib.gis.geos.point import Point
from django.core.files import File
from django.http import HttpResponse, HttpResponseForbidden

logger = logging.getLogger(__name__)


def make_zipfile_bytesio(base_dir):
    zip_file = io.BytesIO()

    with zipfile.ZipFile(zip_file, "w",
                         compression=zipfile.ZIP_DEFLATED) as zf:
        path = os.path.normpath(base_dir)
        if path != os.curdir and path != base_dir:
            zf.write(path, os.path.relpath(path, base_dir))
            logger.info("adding '%s'", path)
        for dirpath, dirnames, filenames in os.walk(base_dir):
            for name in sorted(dirnames):
                path = os.path.normpath(os.path.join(dirpath, name))
                zf.write(path, os.path.relpath(path, base_dir))
                logger.info("adding '%s'", path)
            for name in filenames:
                path = os.path.normpath(os.path.join(dirpath, name))
                if os.path.isfile(path):
                    zf.write(path, os.path.relpath(path, base_dir))
                    logger.info("adding '%s'", path)

    return zip_file


def get_media_response(request, data, permissions=None, headers=None):
    # For compatibility purpose
    content, url = None, None
    if isinstance(data, (io.IOBase, File)):
        content, url = data, data.url
    else:
        # https://docs.djangoproject.com/fr/2.1/ref/request-response/#passing-iterators # noqa
        content, url = open(data['path'], mode='rb'), data['url']

    filetype = magic.from_buffer(content.read(1024), mime=True)
    content.seek(0)

    if isinstance(permissions, list):
        if not set(permissions).intersection(
                request.user.get_all_permissions()):
            return HttpResponseForbidden()

    response = HttpResponse(content_type='application/octet-stream')
    if isinstance(headers, dict):
        for header, value in headers.items():
            response[header] = value

    if settings.MEDIA_ACCEL_REDIRECT:
        response['X-Accel-Redirect'] = f'{url}'
    else:
        response.content = content.read()
        response.content_type = filetype

    return response


class ChunkIterator:

    def __init__(self, iterator, chunksize):
        self.iterator = iterator
        self.chunksize = chunksize

    def __iter__(self):
        return self

    def __next__(self):
        try:
            chunk = []
            for i in range(self.chunksize):
                chunk.append(next(self.iterator))
        finally:
            if chunk:
                return chunk
            else:
                raise StopIteration

    def next(self):
        return self.__next__()


class GeometryDefiner:
    LONGITUDE = 'longitude'
    LATITUDE = 'latitude'

    @staticmethod
    def get_geometry(column_names, row):
        if not isinstance(column_names, dict):
            return None
        if sorted(column_names.keys()) == [GeometryDefiner.LATITUDE,
                                           GeometryDefiner.LONGITUDE]:
            lat_column = column_names.get(GeometryDefiner.LATITUDE)
            long_column = column_names.get(GeometryDefiner.LONGITUDE)
            if all(row.get(column) for column in [long_column, lat_column]):
                x = float(row.get(long_column))
                y = float(row.get(lat_column))
                return Point(x, y)
        return None


class Choices:
    """
    Class under GPL Licence from:
    https://github.com/liberation/django-extended-choices/
    Helper class for choices fields in Django.
    A choice value has three representation (constant name, value and
    string). So Choices takes list of such tuples.
    Here is an example of Choices use:
    >>> CHOICES_ALIGNEMENT = Choices(
    ...     ('BAD', 10, u'bad'),
    ...     ('NEUTRAL', 20, u'neutral'),
    ...     ('CHAOTIC_GOOD', 30, u'chaotic good'),
    ...     ('GOOD', 40, u'good'),
    ... )
    >>> CHOICES_ALIGNEMENT.BAD
    10
    >>> CHOICES_ALIGNEMENT.CHOICES_DICT[30]
    u'chaotic good'
    >>> CHOICES_ALIGNEMENT.REVERTED_CHOICES_DICT[u'good']
    40
    As you can see in the above example usage, Choices objects gets three
    attributes:
    - one attribute built after constant names provided in the tuple (like BAD,
      NEUTRAL etc...)
    - a CHOICES_DICT that match value to string
    - a REVERTED_CHOICES_DICT that match string to value
    If you want to create subset of choices, you can
    use the add_subset method
    This method take a name, and then the constants you want to
    have in this subset:
    >>> CHOICES_ALIGNEMENT.add_subset('WESTERN',('BAD', 'GOOD'))
    >>> CHOICES_ALIGNEMENT.WESTERN
    ((10, u'bad'), (40, u'good'))
    """

    def __init__(self, *raw_choices, **kwargs):
        self._CHOICES = None
        self._CHOICES_DICT = None
        self._REVERTED_CHOICES_DICT = None
        self._VALUE_TO_CONST = None
        self._RAW_CHOICES = None
        self._CONSTS = []
        self._CONST_CHOICES = None
        self.parent = None
        if "parent" in kwargs:  # 'subchoice' mode
            self.parent = kwargs["parent"]
            self._CONSTS = kwargs["CONSTS"]
        else:
            name = kwargs.get('name', 'CHOICES')  # retrocompatibility
            if name != "CHOICES":  # retrocompatibility
                self._RAW_CHOICES = tuple()
                self.add_choices(name, *raw_choices)
            else:
                self._RAW_CHOICES = raw_choices  # default behaviour
                self._make_consts(*raw_choices)

    def _make_consts(self, *raw_choices):
        for choice in raw_choices:
            const, value, string = choice
            if hasattr(self, const):
                raise ValueError("You cannot declare two constants "
                                 "with the same name! %s " % str(choice))
            if value in [getattr(self, c) for c in self._CONSTS]:
                raise ValueError("You cannot declare two constants "
                                 "with the same value! %s " % str(choice))
            setattr(self, const, value)
            self._CONSTS.append(const)

            # retrocompatibilty
            self._CHOICES = None
            self._CHOICES_DICT = None
            self._REVERTED_CHOICES_DICT = None

    def add_choices(self, name="CHOICES", *raw_choices):
        # for retrocompatibility
        # we may have to call _build_choices
        # more than one time and so append the
        # new choices to the already existing ones
        RAW_CHOICES = list(self._RAW_CHOICES)
        self._RAW_CHOICES = tuple(RAW_CHOICES + list(raw_choices))
        self._make_consts(*raw_choices)
        if name != "CHOICES":
            # for retrocompatibility
            # we make a subset with new choices
            constants_for_subset = []
            for choice in raw_choices:
                const, value, string = choice
                constants_for_subset.append(const)
            self.add_subset(name, constants_for_subset)

    def add_subset(self, name, constants):
        if hasattr(self, name):
            raise ValueError("Cannot use %s as a subset name."
                             "It's already an attribute." % name)

        subset = Choices(parent=self, CONSTS=constants)
        setattr(self, name, subset)

        # For retrocompatibility
        setattr(self, '%s_DICT' % name, getattr(subset, "CHOICES_DICT"))
        setattr(self, 'REVERTED_%s_DICT' % name,
                getattr(subset, "REVERTED_CHOICES_DICT"))

    @property
    def RAW_CHOICES(self):
        if self._RAW_CHOICES is None:
            if self.parent:
                self._RAW_CHOICES = tuple((c, k, v) for c, k, v
                                          in self.parent.RAW_CHOICES
                                          if c in self._CONSTS)
            else:
                raise ValueError("Implementation problem : first "
                                 "ancestor should have a _RAW_CHOICES")
        return self._RAW_CHOICES

    @property
    def CHOICES(self):
        """
        Tuple of tuples (value, display_value).
        """
        if self._CHOICES is None:
            self._CHOICES = tuple((k, v) for c, k, v in self.RAW_CHOICES
                                  if c in self._CONSTS)
        return self._CHOICES

    @property
    def CHOICES_DICT(self):
        if self._CHOICES_DICT is None:
            self._CHOICES_DICT = {}
            for c, k, v in self.RAW_CHOICES:
                if c in self._CONSTS:
                    self._CHOICES_DICT[k] = v
        return self._CHOICES_DICT

    @property
    def REVERTED_CHOICES_DICT(self):
        """
        Dict {"display_value": "value"}
        """
        # FIXME: rename in a more friendly name, like STRING_TO_VALUE?
        if self._REVERTED_CHOICES_DICT is None:
            self._REVERTED_CHOICES_DICT = {}
            for c, k, v in self.RAW_CHOICES:
                if c in self._CONSTS:
                    self._REVERTED_CHOICES_DICT[v] = k
        return self._REVERTED_CHOICES_DICT

    @property
    def VALUE_TO_CONST(self):
        """
        Dict {"value": "const"}
        """
        if self._VALUE_TO_CONST is None:
            self._VALUE_TO_CONST = {}
            for c, k, v in self.RAW_CHOICES:
                if c in self._CONSTS:
                    self._VALUE_TO_CONST[k] = c
        return self._VALUE_TO_CONST

    @property
    def CONST_CHOICES(self):
        """
        Tuple of tuples (constant, display_value).
        """
        if self._CONST_CHOICES is None:
            self._CONST_CHOICES = tuple((c, v) for c, k, v in self.RAW_CHOICES
                                        if c in self._CONSTS)
        return self._CONST_CHOICES

    def __contains__(self, item):
        """
        Make smarter to check if a value is valid for a Choices.
        """
        return item in self.CHOICES_DICT

    def __iter__(self):
        return self.CHOICES.__iter__()

    def __eq__(self, other):
        return other == self.CHOICES

    def __ne__(self, other):
        return other != self.CHOICES

    def __len__(self):
        return self.CHOICES.__len__()

    def __add__(self, item):
        """
        Needed to make MY_CHOICES + OTHER_CHOICE
        """
        if not isinstance(item, (Choices, tuple)):
            raise ValueError("This operand could only by evaluated "
                             "with Choices or tuple instances. "
                             "Got %s instead." % type(item))
        return self.CHOICES + tuple(item)

    def __repr__(self):
        return self.CHOICES.__repr__()



import logging
import types
from datetime import date, timedelta

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group
from django.core.mail import send_mail
from django.db.models import Model
from django.forms.models import model_to_dict
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from simpleeval import EvalWithCompoundTypes, simple_eval

from terracommon.events.signals import event

from . import funcs

logger = logging.getLogger(__name__)


class AbstractHandler(object):

    default_settings = {
        'condition': 'True',
    }
    default_args = {
        'instance': None,
    }

    def __init__(self, event, settings, **kwargs):
        """
        :param event: Is the event name raised
        :param settings: Is a dict from settings of EventHandler model. It
                         override default Handler settings.
        :param **kwargs: Are all extra parameters provided by the event
                         sender. All those arguments are provided to all
                         eval or template.

        """

        self.event = event
        self.handler_settings = settings or {}
        self.args = {**self.default_args, **kwargs}

    @property
    def settings(self):
        return {**self.default_settings, **self.handler_settings}

    def valid_condition(self):
        return simple_eval(
            self.settings['condition'],
            names=self.vars,
            functions=self.functions,
            )

    @cached_property
    def vars(self):
        attrs = {k: str(v) for k, v in self.args.items()}
        attrs.update({
            'settings': self.settings,
            'event': self.event,
            'instance': self.serialized_instance,
            'front_url': settings.FRONT_URL,
            'hostname': settings.HOSTNAME,
        })

        return attrs

    @cached_property
    def serialized_instance(self):
        instance = self.args.get('instance')
        if isinstance(instance, Model):
            return model_to_dict(instance)
        try:
            return vars(instance)
        except TypeError:
            return {}

    @cached_property
    def functions(self):
        functions = {
            f: getattr(funcs, f)
            for f in dir(funcs)
            if isinstance(getattr(funcs, f), types.FunctionType)
        }

        functions.update({
            'min': min,
            'max': max,
            'any': any,
            'all': all,
            'sum': sum,
        })

        return functions

    def __call__(self):
        raise NotImplementedError


class SendEmailHandler(AbstractHandler):
    """
    This handler send an email to the list of users returned by the «emails»
    interpreted settings.
    subject_tpl and body_tpl are formatted with python .format() method.
    """

    default_settings = {
        'condition': 'True',
        'from_email': settings.DEFAULT_FROM_EMAIL,
        'recipients': "[user['email'], ]",
        'subject_tpl': "Hello world {user[email]}",
        'body_tpl': "Dear, your properties {user[properties]}"
    }

    def __call__(self):
        s = EvalWithCompoundTypes(
            names=self.vars,
            functions=self.functions
            )
        recipients = s.eval(self.settings['recipients'])

        for recipient in recipients:
            recipient_data = self._get_recipient_data(recipient)

            subject = self.settings['subject_tpl'].format(
                recipient=recipient_data,
                **self.vars,)
            body = self.settings['body_tpl'].format(
                recipient=recipient_data,
                **self.vars,)
            to_email = getattr(recipient_data, 'email', recipient)

            if to_email:
                send_mail(
                    subject,
                    body,
                    self.settings['from_email'],
                    [to_email, ],
                    fail_silently=True,
                    )
            else:
                logger.error('No destination e-mail could be found for %s',
                             recipient)

    @cached_property
    def vars(self):
        vars = super().vars
        if isinstance(self.args['user'], get_user_model()):
            vars.update({
                'user': {
                    'email': self.args['user'].email,
                    'properties': self.args['user'].properties,
                },
            })
        return vars

    def _get_recipient_data(self, o):
        if isinstance(o, get_user_model()):
            return o
        try:
            user = get_user_model().objects.get(email=o)
            return {
                'email': user.email,
                'properties': user.properties,
            }
        except get_user_model().DoesNotExist:
            return None


class TimeDeltaHandler(AbstractHandler):
    '''Set the ```field``` element of the instance as a timedelta from today.
    It's possible to set a field content, as it's possible to set a property
    content.
    Available field from UserRequest are ```['expiry', 'properties', ]```
    '''

    default_settings = {
        'condition': 'True',
        'daysdelta': 0,
        'field': 'expiry',
    }

    def __call__(self):
        field_path = self.settings['field'].split('.')

        if field_path[0] not in ['expiry', 'properties']:
            return

        self._set_value(field_path, self.args['instance'])
        self.args['instance'].save()

    def _set_value(self, field_path, attr):
        if len(field_path) > 1:
            field_name = field_path.pop(0)

            modified_attr = getattr(attr, field_name, {})
            modified_attr = self._set_value(field_path, modified_attr)

            self._setattr(attr, field_name, modified_attr)

        else:
            self._setattr(attr, field_path.pop(),
                          str(date.today() + self._get_timedelta()))

        return attr

    def _setattr(self, attr, key, value):
        try:
            setattr(attr, key, value)
        except AttributeError:
            attr[key] = value

    def _get_timedelta(self):
        return timedelta(days=int(self.settings['daysdelta']))


class SendNotificationHandler(AbstractHandler):

    default_settings = {
        'condition': 'True',
        'level': 'info',
        'message': "New notifications received",
        'event_code': 'default_notification',
    }

    def __call__(self):
        message = self.settings['message'].format(**self.vars)
        uuid = (self.args['instance'].uuid
                if hasattr(self.args['instance'], 'uuid') else None)

        self.args['user'].notifications.create(
            level=self.settings.get('level'),
            message=message,
            event_code=self.settings.get('event_code'),
            identifier=self.args['instance'].pk,
            uuid=uuid,
        )


class SetGroupHandler(AbstractHandler):
    default_settings = {
        'condition': 'True',
        'group': None,
        'userfield': None,
    }

    def __call__(self):
        group = Group.objects.get(name=self.settings['group'])
        self.args[self.settings['userfield']].groups.add(group)


class ModelValueHandler(AbstractHandler):
    """
    Retrieves model records that validates the query.
    The query values are evaluated before being passed to the ORM.
    Triggers actions for each recovered record.
    """

    default_settings = {
        'condition': 'True',
        'model': None,
        'query': {},
        'actions': None,
    }

    def __call__(self):
        instances = self._get_queryset().all()
        actions = self.settings['actions']
        for instance in instances:
            for action in actions:
                kwargs = {'user': AnonymousUser()}
                kwargs.update(**action.get('kwargs', {}))
                kwargs.update({'action': action['action'],
                               'instance': instance, })
                event.send(self.__class__, **kwargs)

    def _get_queryset(self):
        model = import_string(self.settings['model'])
        kwargs = {
            k: simple_eval(v,
                           names=self.vars,
                           functions=self.functions, )
            for k, v in self.settings['query'].items()
        }
        return model.objects.filter(**kwargs)

    @cached_property
    def serialized_instance(self):
        return None

    @cached_property
    def functions(self):
        functions = super().functions
        functions.update({
            'date': date,
            'timedelta': timedelta,
        })
        return functions



import csv
import logging
from copy import deepcopy

logger = logging.getLogger(__name__)


class CustomCsvBuilder:
    """ Build csv file from dict/list data """

    def __init__(self, data):
        self.data = data

    def build_header(self, data, parent_key=None):
        header = []

        if isinstance(data, dict):
            for colname, colvalue in data.items():
                full_colname = colname
                if parent_key:
                    full_colname = (parent_key
                                    if colname in parent_key
                                    else '_'.join((parent_key, colname)))

                partial_header = self.build_header(
                    colvalue,
                    parent_key=full_colname,
                )

                header += partial_header

        elif isinstance(data, list):
            if parent_key:
                header.append(parent_key)
            for val in data:
                header += self.build_header(val, parent_key=parent_key)

        else:
            if parent_key:
                header.append(parent_key)
        return set(header)

    def flatten_datatree(self, tree, parent='', sep='_'):
        flatten_tree = []
        if isinstance(tree, dict):
            self._tree_flattend(flatten_tree, tree, parent, sep)
        elif isinstance(tree, list):
            self._list_flattend(flatten_tree, tree, parent, sep)
        else:
            flatten_tree.append({parent: tree})
        return flatten_tree

    def _list_flattend(self, flatten_tree, tree, parent='', sep='_'):
        for i, sibling in enumerate(tree):
            flatten_sibling = self.flatten_datatree(sibling, parent=parent)
            flatten_tree.extend(flatten_sibling)

    def _tree_flattend(self, flatten_tree, tree, parent='', sep='_'):
        for node, child in tree.items():
            parent_node = node if not parent else sep.join((parent, node))
            flatten_child = self.flatten_datatree(child,
                                                  parent=parent_node)

            if not flatten_tree:
                flatten_tree.extend(flatten_child)
            else:
                for i, sub_child in enumerate(flatten_child):
                    for sub_tree in flatten_tree:
                        if sub_child.keys() <= sub_tree.keys():
                            subtree_copy = deepcopy(sub_tree)
                            subtree_copy.update(sub_child)
                            flatten_tree.append(subtree_copy)
                            break
                    else:
                        try:
                            flatten_tree[i].update(sub_child)
                        except IndexError:
                            flatten_tree.append(sub_child)

    def create_csv(self, output):
        self.csv = output

        header = self.build_header(self.data)
        body = self.flatten_datatree(self.data)

        writer = csv.DictWriter(self.csv, fieldnames=header)

        writer.writeheader()
        for element in body:
            writer.writerow(element)
        output.seek(0)



# Generated by Django 2.0.13 on 2019-04-01 11:39

import django.contrib.postgres.fields.jsonb
from django.db import migrations
import terracommon.terra.validators


class Migration(migrations.Migration):

    dependencies = [
        ('terra', '0019_merge_20190329_1056'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='layer',
            options={'ordering': ['id'], 'permissions': (('can_update_features_properties', 'Is able update geometries properties'), ('can_export_layers', 'Is able to export layers'), ('can_import_layers', 'Is able to import layers'))},
        ),
        migrations.AlterField(
            model_name='layer',
            name='schema',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, validators=[terracommon.terra.validators.validate_json_schema]),
        ),
    ]



from django.contrib.gis.db.models import GeometryField
from django.contrib.gis.geos import Polygon
from django.db.models.functions import Cast

from . import EPSG_3857
from .funcs import ST_Extent, ST_Transform


class SIGTools(object):
    """
    This class collects all the sig functions  to use in TileVect, GeomConfig
    """

    @classmethod
    def get_extent_of_layer(cls, layer):
        """
        Outputs length of the smaller size of the features' bbox
        Returns 0 if the bbox is a Point
        """

        min_extent_features = float(0)

        # output might be a Point if single Point feature
        query = cls.get_bbox_layer(layer)

        if isinstance(query['extent'], Polygon):

            x1 = query['extent'].coords[0][0][0]
            x2 = query['extent'].coords[0][2][0]
            y1 = query['extent'].coords[0][0][1]
            y2 = query['extent'].coords[0][2][1]

            min_extent_features = min(float(abs(x2-x1)), float(abs(y2-y1)))

        return min_extent_features

    @classmethod
    def get_bbox_layer(cls, layer):
        """
        Method giving the bbox of a layer,
        under the form of a Polygon (if > 0)
        OUTPUT query # Polygon (bbox) with 'extent' attribute
        """
        features = layer.features.all()

        query = features.annotate(
            geom3857=Cast(ST_Transform('geom', EPSG_3857),
                          GeometryField(srid=EPSG_3857))
            ).aggregate(
            extent=Cast(ST_Extent('geom3857'), GeometryField(srid=0))
            )

        return query



import json

from django.contrib.gis.geos.geometry import GEOSGeometry
from django.test import TestCase
from django.urls import reverse
from rest_framework.status import HTTP_200_OK

from terracommon.terra.tests.factories import (FeatureFactory, LayerFactory,
                                               UserFactory)


class FeaturesListViewTest(TestCase):
    fake_geometry = {
        "type": "Point",
        "coordinates": [
          2.,
          45.
        ]
    }
    intersect_geometry = {
        "type": "LineString",
        "coordinates": [
          [
            1.3839340209960938,
            43.602521593464054
          ],
          [
            1.4869308471679688,
            43.60376465190968
          ]
        ]
    }
    intersect_ref_geometry = {
        "type": "LineString",
        "coordinates": [
            [
                1.440925598144531,
                43.64750394449096
            ],
            [
                1.440582275390625,
                43.574421623084234
            ]
        ]
    }
    fake_linestring = {
        "type": "LineString",
        "coordinates": [
            [
                1.3839340209960938,
                43.602521593464054
            ],
        ]
    }
    fake_polygon = {
        "type": "Polygon",
        "coordinates": [
            [
                [
                    1.3839340209960938,
                    43.602521593464054
                ],
                [
                    1.440582275390625,
                    43.574421623084234
                ]
            ]
        ]
    }

    group_name = 'mygroup'

    def setUp(self):
        self.layer = LayerFactory.create(group=self.group_name,
                                         add_features=5)

        self.user = UserFactory()
        self.client.force_login(self.user)

    def test_features_filter_by_properties(self):
        layer = LayerFactory(group=self.group_name)
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'number': 1, 'text': 'bar'},
        )
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'number': 1, 'text': 'foo'},
        )
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'number': 2, 'text': 'foo'},
        )
        response = self.client.get(
            reverse('terra:feature-list', kwargs={'layer': layer.pk}),
            {'properties__number': 1, 'properties__text': 'foo'},
        )
        self.assertEqual(response.status_code, HTTP_200_OK)
        json_response = response.json()
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(len(json_response['results']), 1)

    def test_features_filter_by_properties_with_wrong_field(self):
        layer = LayerFactory(group=self.group_name)
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'number': 1},
        )
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'number': 2},
        )
        response = self.client.get(
            reverse('terra:feature-list', kwargs={'layer': layer.pk}),
            {'properties__wrongfield': 'wrong value'},
        )
        self.assertEqual(response.status_code, HTTP_200_OK)

        json_response = response.json()
        self.assertEqual(json_response['count'], 0)
        self.assertEqual(len(json_response['results']), 0)

    def test_features_filter_by_properties_with_several_int_field(self):
        layer = LayerFactory()
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'number': 2, 'digit': 42},
        )
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'number': 1, 'digit': 42},
        )
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'number': 1, 'digit': 34},
        )
        response = self.client.get(
            reverse('terra:feature-list', kwargs={'layer': layer.pk}),
            {'properties__number': 1, 'properties__digit': 42},
        )
        self.assertEqual(response.status_code, HTTP_200_OK)
        json_response = response.json()
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(len(json_response['results']), 1)

    def test_features_filter_by_properties_with_several_string_field(self):
        layer = LayerFactory()
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'text': 'foobar', 'sentence': 'foobar is here'},
        )
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'text': 'foo', 'sentence': 'foobar is missing'},
        )
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'text': 'foobar', 'sentence': 'foobar is here'},
        )
        response = self.client.get(
            reverse('terra:feature-list', kwargs={'layer': layer.pk}),
            {
                'properties__text': 'foobar',
                'properties__sentence': 'foobar is here'
            }
        )
        self.assertEqual(response.status_code, HTTP_200_OK)
        json_response = response.json()
        self.assertEqual(json_response['count'], 2)
        self.assertEqual(len(json_response['results']), 2)

    def test_feature_from_layer_name(self):
        layer = LayerFactory()
        feature = FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.fake_geometry)),
            properties={'text': 'foobar', 'sentence': 'foobar is here'},
        )

        response = self.client.get(
            reverse(
                'terra:feature-detail',
                kwargs={'layer': str(layer.name),
                        'identifier': str(feature.identifier)}
            ),
        )
        self.assertEqual(response.status_code, HTTP_200_OK)



# Generated by Django 2.0.9 on 2018-10-24 16:09

import django.db.models.deletion
from django.db import migrations, models

import terracommon.datastore.models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('datastore', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='RelatedDocument',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=255)),
                ('object_id', models.PositiveIntegerField()),
                ('document', models.FileField(upload_to=terracommon.datastore.models.related_document_path)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='contenttypes.ContentType')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='relateddocument',
            unique_together={('key', 'content_type', 'object_id')},
        ),
    ]



from rest_framework import routers

from .views import CommentViewSet, RequestViewSet

app_name = 'trrequests'

router = routers.SimpleRouter()

router.register(r'userrequest', RequestViewSet, base_name='request')
router.register(r'userrequest/(?P<request_pk>\d+)/comment',
                CommentViewSet, base_name='comment')

urlpatterns = router.urls



import base64
from datetime import timedelta
from tempfile import NamedTemporaryFile

from django.utils.dateparse import parse_datetime
from docx.shared import Mm
from docxtpl import InlineImage
from jinja2 import contextfilter

from terracommon.datastore.models import DataStore


def timedelta_filter(date_value, **kwargs):
    """ custom filter that will add a positive or negative value, timedelta
        to the day of a date in string format """

    current_date = parse_datetime(date_value)
    return (current_date - timedelta(**kwargs))


def translate_filter(value, datastorekey=''):
    """ Custom filter that will return the translated data
        from datastore correspondence table """

    if not datastorekey:
        return value
    correspondences = DataStore.objects.get(key=datastorekey)
    return correspondences.value.get(value, value)


def todate_filter(value):
    return parse_datetime(value).date()


@contextfilter
def b64_to_inlineimage(context, value):
    b64str = value.split(',', 1)[1]
    decoded = base64.b64decode(b64str)
    tmp_f = NamedTemporaryFile(mode='wb', dir=context['tmpdir'], delete=False)
    tmp_f.write(decoded)
    tmp_f.seek(0)
    return InlineImage(context['tpl'], tmp_f.name, width=Mm(40))



import os

MEDIA_ACCEL_REDIRECT = os.getenv('MEDIA_ACCEL_REDIRECT', default="False") == "True"
HOSTNAME = os.environ.get('HOSTNAME', '')

TERRA_TILES_HOSTNAMES = [
    HOSTNAME,
]

ALLOWED_HOSTS = []

DEBUG = True
SWAGGER_ENABLED = False

MAX_TILE_ZOOM = 15
MIN_TILE_ZOOM = 10

INTERNAL_GEOMETRY_SRID = 4326



# Generated by Django 2.0.9 on 2018-11-20 10:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('datastore', '0002_auto_20181024_1609'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='datastore',
            options={'ordering': ['key']},
        ),
        migrations.AlterModelOptions(
            name='datastorepermission',
            options={'ordering': ['prefix'], 'permissions': (('can_read_datastore', "Is able to read all datastore's elements"), ('can_readwrite_datastore', 'Is able to write in datastore'))},
        ),
        migrations.AlterModelOptions(
            name='relateddocument',
            options={'ordering': ['key']},
        ),
    ]



from django.contrib.auth.base_user import BaseUserManager
from django.contrib.contenttypes.models import ContentType
from django.db.models import Manager, ObjectDoesNotExist


class TerraUserManager(BaseUserManager):

    def _create_user(self, email, password, **extra_fields):
        email = self.normalize_email(email)
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email, password, **extra_fields):
        extra_fields.setdefault('is_staff', False)
        extra_fields.setdefault('is_superuser', False)
        return self._create_user(email, password, **extra_fields)

    def create_superuser(self, email, password, **extra_fields):
        extra_fields.setdefault('is_staff', True)
        extra_fields.setdefault('is_superuser', True)

        return self._create_user(email, password, **extra_fields)


class ReadModelManager(Manager):
    def get_user_read(self, user, obj):
        contenttype = ContentType.objects.get_for_model(obj.__class__)
        try:
            return self.get(
                user=user,
                contenttype=contenttype,
                identifier=obj.pk)
        except ObjectDoesNotExist:
            return None

    def read_object(self, user, obj):
        read = self.get_user_read(user, obj)
        if read:
            read.read_instance()
        else:
            read = self.create(
                user=user,
                model_object=obj
            )

        return read



# Generated by Django 2.0.13 on 2019-04-17 16:30

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('accounts', '0004_auto_20181120_1059'),
    ]

    operations = [
        migrations.AlterField(
            model_name='terrauser',
            name='uuid',
            field=models.CharField(default=uuid.uuid4, max_length=255, unique=True, verbose_name='unique identifier'),
        ),
    ]



import argparse
import csv
import sys

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.core.management import BaseCommand
from django.db import transaction
from django.utils.translation import ugettext as _

UserModel = get_user_model()


class Command(BaseCommand):
    help = _('Import users from csv to database.')

    def add_arguments(self, parser):
        parser.add_argument('-cd', '--delimiter',
                            action='store',
                            dest='delimiter',
                            default=';',
                            required=False,
                            help=_('Specify CSV delimiter')
                            )
        parser.add_argument('-cq', '--quotechar',
                            action='store',
                            dest='quotechar',
                            default='"',
                            required=False,
                            help=_('Specify CSV quotechar')
                            )
        parser.add_argument('-cs', '--source',
                            dest='source',
                            type=argparse.FileType('r',
                                                   encoding='iso-8859-15'),
                            default=sys.stdin,
                            required=True,
                            help=_('Specify CSV path'),
                            )
        parser.add_argument('-u', '--username-field',
                            required=True,
                            action='store',
                            dest='user_field',
                            help=_("Define the column used as userfield, "
                                   "others are put in properties")
                            )
        parser.add_argument('-p', '--password-field',
                            required=False,
                            action='store',
                            dest='password_field',
                            help=_("Define the column used as userfield, "
                                   "others are put in properties")
                            )
        parser.add_argument('-g', '--group',
                            required=False,
                            action='append',
                            dest='group',
                            help=_("Default group of newly created users")
                            )

    @transaction.atomic
    def handle(self, *args, **options):
        username, password = (options.get('user_field'),
                              options.get('password_field'))

        reader = csv.DictReader(options.get('source'),
                                delimiter=options.get('delimiter'),
                                quotechar=options.get('quotechar'))

        fields = [fieldname
                  for fieldname in reader.fieldnames
                  if fieldname not in ['', username, password]]

        groups = Group.objects.filter(name__in=options.get('group'))

        for row in reader:
            user = UserModel.objects.create(
                **{
                    UserModel.USERNAME_FIELD: row[username],
                    'properties': {field: row.get(field) for field in fields},
                }
            )
            if password:
                user.set_password(row[password])
            else:
                user.set_unusable_password()

            user.groups.add(*groups)



# Generated by Django 2.0.8 on 2018-09-11 08:18

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('accounts', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='terrauser',
            options={'permissions': (('can_manage_users', 'Is able create, delete, update users'),)},
        ),
    ]



from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.postgres.fields import JSONField
from django.db import models
from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from terracommon.accounts.mixins import ReadableModelMixin
from terracommon.core.mixins import BaseUpdatableModel
from terracommon.datastore.models import RelatedDocument
from terracommon.document_generator.models import DownloadableDocument
from terracommon.terra.models import Layer

from .helpers import rename_comment_attachment


class UserRequest(BaseUpdatableModel, ReadableModelMixin):
    owner = models.ForeignKey(settings.AUTH_USER_MODEL,
                              on_delete=models.PROTECT,
                              related_name='userrequests')
    layer = models.ForeignKey(Layer,
                              on_delete=models.PROTECT,
                              related_name='userrequests')
    expiry = models.DateField(default=None, null=True)
    state = models.IntegerField(default=settings.STATES.DRAFT)
    reviewers = models.ManyToManyField(settings.AUTH_USER_MODEL,
                                       blank=True,
                                       related_name='to_review')
    properties = JSONField(default=dict, blank=True)
    downloadable = GenericRelation(DownloadableDocument)
    documents = GenericRelation(RelatedDocument)

    def get_comments_for_user(self, user):
        query = self.comments.all()
        if not user:
            return query.none()

        filter = Q()

        # exclude comments if the user have no permission
        if not user.has_perm(
                'trrequests.can_internal_comment_requests'):
            filter |= Q(is_internal=True)

        if (not user.has_perm('trrequests.can_comment_requests') and
                not user.has_perm('trrequests.can_read_comment_requests')):
            filter |= Q(is_internal=False)

        return query.exclude(filter)

    def get_serializer(self):
        # Exceptionnally,
        # to avoid circular dependencies between model and serializer
        from .serializers import UserRequestSerializer

        return UserRequestSerializer

    def get_pdf_serializer(self):
        # Avoid circular dependencies
        from .serializers import UserRequestPDFSerializer
        return UserRequestPDFSerializer

    class Meta:
        ordering = ['id']
        permissions = (
            ('can_create_requests', 'Is able to create a new requests'),
            ('can_read_self_requests', 'Is able to get own requests'),
            ('can_read_all_requests', 'Is able to get all requests'),
            ('can_comment_requests', 'Is able to comment an user request'),
            ('can_internal_comment_requests',
             'Is able to add comments not visible by users'),
            ('can_read_comment_requests',
             'Is allowed to read only non-internal comments'),
            ('can_change_state_requests',
             'Is authorized to change the request state'),
            ('can_download_all_pdf',
             'Is able to download a pdf document')
        )


class Comment(BaseUpdatableModel):
    owner = models.ForeignKey(settings.AUTH_USER_MODEL,
                              on_delete=models.PROTECT)
    userrequest = models.ForeignKey(UserRequest,
                                    on_delete=models.PROTECT,
                                    related_name="comments")
    layer = models.ForeignKey(Layer,
                              null=True,
                              on_delete=models.PROTECT)
    properties = JSONField(default=dict, blank=True)
    is_internal = models.BooleanField(default=False)
    attachment = models.FileField(upload_to=rename_comment_attachment,
                                  blank=True)
    filename = models.CharField(max_length=255,
                                editable=False,
                                blank=True,
                                help_text=_('Initial name of the attachment'))

    def save(self, *args, **kwargs):
        if not self.attachment.name.startswith('userrequests/'):
            # Save the name of the new file before upload_to move & rename it
            self.filename = self.attachment.name
        super().save(*args, **kwargs)

    class Meta:
        ordering = ['id']



# Generated by Django 2.0.9 on 2018-11-20 10:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('document_generator', '0004_auto_20181012_0755'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='documenttemplate',
            options={'ordering': ['id'], 'permissions': (('can_upload_template', 'Is allowed to upload a template'), ('can_update_template', 'Is allowed to update a template'), ('can_delete_template', 'Is allowed to delete a template'))},
        ),
        migrations.AlterModelOptions(
            name='downloadabledocument',
            options={'ordering': ['id']},
        ),
    ]



from io import StringIO

from django.core.cache import cache
from django.core.management import call_command
from django.db import connection
from django.test import TestCase, override_settings

from terracommon.terra.tests.factories import LayerFactory
from terracommon.terra.tiles.helpers import VectorTile, get_cache_version


@override_settings(DEBUG=True, CACHES={
    'default': {
        'BACKEND': ('django.core.cache.backends'
                    '.locmem.LocMemCache')
    }})
class FillTilesCacheTestCase(TestCase):
    group_name = 'mygroup'

    def setUp(self):
        self.layer = LayerFactory(group=self.group_name, name="layerLine")

        self.layer.from_geojson(
            geojson_data='''
            {
            "type": "FeatureCollection",
            "features": [
                {
                "type": "Feature",
                "properties": {
                    "foo": "bar",
                    "baba": "fifi"
                },
                "geometry": {
                    "type": "LineString",
                    "coordinates": [
                    [
                        1.3700294494628906,
                        43.603640347220924
                    ],
                    [
                        1.2984466552734375,
                        43.57902295875415
                    ]
                    ]
                }
                }
            ]
            }
        ''')

    def test_update_topology_routing_fail(self):
        tile = VectorTile(self.layer)
        features = self.layer.features.all()

        cache_version = get_cache_version(self.layer)

        x, y, z = 515, 373, 10
        pixel_buffer, features_filter, properties_filter, features_limit = 4, None, None, 10000

        query_count_before = len(connection.queries)

        call_command('fill_tiles_cache', stdout=StringIO())

        query_count_after = len(connection.queries)

        self.assertLess(query_count_before, query_count_after)

        tile.get_tile(
            x, y, z,
            pixel_buffer, features_filter, properties_filter, features_limit,
            features)

        self.assertIsNotNone(
            cache.get(
                tile.get_tile_cache_key(x, y, z, pixel_buffer, features_filter, properties_filter, features_limit),
                version=cache_version,
            )
        )

        self.assertEqual(len(connection.queries), query_count_after + 1)



from django.conf.urls import url
from django.urls import path
from rest_framework import routers

from .views import (SettingsView, UserChangePasswordView, UserInformationsView,
                    UserProfileView, UserRegisterView, UserSetPasswordView,
                    UserViewSet)

app_name = 'accounts'

router = routers.SimpleRouter()
router.register(r'user', UserViewSet, base_name='user')
urlpatterns = router.urls

urlpatterns += [
    path(r'settings/', SettingsView.as_view(), name='settings'),
    url(r'^auth/user/', UserInformationsView.as_view()),
    url(r'^accounts/user/$', UserProfileView.as_view(), name='profile'),
    url(r'^accounts/register/$', UserRegisterView.as_view(), name='register'),
    url((r'^accounts/change-password/reset/(?P<uidb64>[0-9A-Za-z_\-]+)/'
         r'(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,20})/$'),
        UserSetPasswordView.as_view(), name='reset-password'),
    url(r'^accounts/change-password/reset/',
        UserChangePasswordView.as_view(), name='new-password'),
]



from rest_framework import status, viewsets
from rest_framework.decorators import list_route
from rest_framework.response import Response

from .serializers import UserNotificationSerializer


class NotificationViewSet(viewsets.ReadOnlyModelViewSet):
    serializer_class = UserNotificationSerializer

    def get_queryset(self, *args, **kwargs):
        return self.request.user.notifications.all()

    @list_route(methods=['get'])
    def read_all(self, *args, **kwargs):
        self.request.user.notifications.read_all()
        return Response(status.HTTP_200_OK)



import jsonschema
from django.core.exceptions import ValidationError
from jsonschema.validators import validator_for


def validate_json_schema(value):
    """
    Validate json schema
    """
    try:
        if value:
            # check only if schema defined
            cls = validator_for(value)
            cls.check_schema(value)
    except Exception as e:
        raise ValidationError(message=e.message)

    return value


def validate_json_schema_data(value, schema):
    """
    Validate data according json schema
    """
    try:
        # check result schema
        if value and schema:
            properties = schema.get('properties').keys()
            unexpected_properties = value.keys() - properties
            if unexpected_properties:
                # value key(s) not in expected properties
                raise ValidationError(message=f"{unexpected_properties} not in schema properties")
            jsonschema.validate(value, schema)
    except jsonschema.exceptions.ValidationError as e:
        raise ValidationError(message=e.message)

    return value


def validate_geom_type(layer_geom_type_id, feature_geom_type_id):
    if isinstance(layer_geom_type_id, int):
        if layer_geom_type_id != feature_geom_type_id:
            raise ValidationError(message=f'Geometry type is not the same on the layer')
    return feature_geom_type_id



from ..core.settings import *

DEFAULT_FILE_STORAGE = 'storages.backends.s3boto3.S3Boto3Storage'

INSTALLED_APPS += (
    'versatileimagefield',
)

VERSATILEIMAGEFIELD_RENDITION_KEY_SETS = {
    'tropp': [
        ('original', 'url'),
        ('full', 'thumbnail__1500x1125'),
        ('list', 'thumbnail__300x225'),
        ('thumbnail', 'thumbnail__180x120'),
    ]
}

TROPP_BASE_LAYER_NAME = 'Base opp layer'

TROPP_PICTURES_STATES_WORKFLOW = False

TROPP_VIEWPOINT_PROPERTIES_SET = {
    'pdf': {
        ('camera', 'Appareil photo'),
    },
    'form': {},
    'filter': {},
}

TROPP_FEATURES_PROPERTIES_FROM_VIEWPOINT = [
    'commune',
]

TROPP_SEARCHABLE_PROPERTIES = {
    'cities': {
        'json_key': 'commune',
        'type': 'single',
    },
    'themes': {
        'json_key': 'themes',
        'type': 'many',
    },
    'road': {
        'json_key': 'voie',
        'type': 'text',
    },
    'site': {
        'json_key': 'site',
        'type': 'text',
    },
}



from rest_framework import viewsets
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.status import HTTP_409_CONFLICT

from terracommon.core.renderers import CSVRenderer

from .models import DataStore
from .permissions import IsAuthenticatedAndDataStoreAllowed
from .serializers import DataStoreSerializer


class DataStoreViewSet(viewsets.ModelViewSet):
    renderer_classes = (JSONRenderer, CSVRenderer, )

    permission_classes = (IsAuthenticatedAndDataStoreAllowed, )
    serializer_class = DataStoreSerializer
    lookup_field = 'key'
    lookup_value_regex = '[^/]+'

    def get_queryset(self):
        return DataStore.objects.get_datastores_for_user(self.request.user)

    def update(self, request, key=None):
        obj = self.get_object()

        serializer = self.get_serializer(obj, data={'value': request.data})

        serializer.is_valid(raise_exception=True)
        serializer.save()

        return Response(serializer.data.get('value'))

    def post(self, request, *args, **kwargs):
        attrs = {self.lookup_field: kwargs.get(self.lookup_field)}

        obj = DataStore(**attrs)
        serializer = self.get_serializer(obj, data={'value': request.data})
        self.check_object_permissions(request, obj)

        if DataStore.objects.filter(**attrs).exists():
            return Response(status=HTTP_409_CONFLICT)

        serializer.is_valid(raise_exception=True)
        serializer.save()

        return Response(serializer.data.get('value'))



from django.contrib.postgres.fields import JSONField
from django.db import models


class EventHandler(models.Model):
    action = models.CharField(max_length=255, blank=False, null=False)
    handler = models.CharField(max_length=255, blank=False, null=False)
    settings = JSONField(default=dict)
    priority = models.PositiveIntegerField(default=10)

    class Meta:
        ordering = ['id']



import uuid

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models

UserModel = get_user_model()


class DocumentTemplate(models.Model):
    name = models.CharField(max_length=50)
    documenttemplate = models.FileField(upload_to='templates/%Y/%m/')
    uid = models.CharField(max_length=256, default=uuid.uuid4)

    class Meta:
        ordering = ['id']
        permissions = (
            ('can_upload_template', 'Is allowed to upload a template'),
            ('can_update_template', 'Is allowed to update a template'),
            ('can_delete_template', 'Is allowed to delete a template'),
        )


class DownloadableDocument(models.Model):
    user = models.ForeignKey(UserModel, on_delete=models.PROTECT)
    document = models.ForeignKey(DocumentTemplate, on_delete=models.PROTECT)

    content_type = models.ForeignKey(ContentType, on_delete=models.PROTECT)
    object_id = models.PositiveIntegerField()
    linked_object = GenericForeignKey('content_type', 'object_id')

    class Meta:
        ordering = ['id']



import operator
from collections import OrderedDict
from functools import reduce

import coreapi
import coreschema
from django.contrib.postgres.fields.jsonb import KeyTransform
from django.db.models import Prefetch
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from rest_framework import permissions, viewsets
from rest_framework.decorators import action
from rest_framework.filters import SearchFilter
from rest_framework.generics import RetrieveAPIView
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response

from ..core.filters import DateFilterBackend, SchemaAwareDjangoFilterBackend
from ..core.renderers import PdfRenderer, ZipRenderer
from .filters import CampaignFilterBackend, JsonFilterBackend
from .serializers import *


class ViewpointPdf(RetrieveAPIView):
    """
    Return a pdf representation of the given viewpoint by its id.
    """
    queryset = Viewpoint.objects.all()
    template_name = 'tropp/viewpoint_pdf.html'
    renderer_classes = (PdfRenderer,)
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)

    @method_decorator(cache_page(60 * 5))
    def get(self, request, *args, **kwargs):
        properties_set = settings.TROPP_VIEWPOINT_PROPERTIES_SET['pdf']
        return Response({
            'viewpoint': self.get_object(),
            'properties_set': properties_set,
        })


class ViewpointZipPictures(RetrieveAPIView):
    """
    Return a zip archive of all pictures
    """
    queryset = Viewpoint.objects.all()
    renderer_classes = (ZipRenderer,)
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)

    @method_decorator(cache_page(60 * 5))
    def get(self, request, *args, **kwargs):
        qs = self.get_object().pictures.filter(
            state__gte=settings.STATES.ACCEPTED,
        ).only('file')
        return Response([p.file for p in qs])


class RestPageNumberPagination(PageNumberPagination):
    page_size_query_param = 'page_size'

    def get_paginated_response(self, data):
        page = self.page

        next_page = page.next_page_number() if page.has_next() else None
        previous_page = (page.previous_page_number() if page.has_previous()
                         else None)

        return Response(OrderedDict([
            ('count', page.paginator.count),
            ('num_pages', page.paginator.num_pages),
            ('next', next_page),
            ('previous', previous_page),
            ('page_size', self.get_page_size(self.request)),
            ('results', data),
        ]))


class ViewpointViewSet(viewsets.ModelViewSet):
    serializer_class = ViewpointSerializerWithPicture
    permission_classes = [
        permissions.DjangoModelPermissionsOrAnonReadOnly,
    ]
    filter_backends = (
        SearchFilter,
        SchemaAwareDjangoFilterBackend,
        DateFilterBackend,
        JsonFilterBackend,
    )
    filter_fields_schema = [
        coreapi.Field(
            name='pictures__id',
            required=False,
            location='query',
            schema=coreschema.Integer(
                title="Picture id",
                description="Picture id to filter on",
            ),
        ),
        coreapi.Field(
            name='pictures__owner__uuid',
            required=False,
            location='query',
            schema=coreschema.Integer(
                title="Photographer uuid",
                description="Photographer uuid to filter on",
            ),
        ),
    ]
    filter_fields = ['pictures']
    search_fields = ('label', )
    date_search_field = 'pictures__date__date'
    pagination_class = RestPageNumberPagination

    def filter_queryset(self, queryset):
        # We must reorder the queryset here because initial filtering in
        # viewpoint model is not done right see
        # https://github.com/encode/django-rest-framework/issues/1717
        return super().filter_queryset(queryset).order_by('-created_at')

    def get_queryset(self):
        qs = Viewpoint.objects.with_accepted_pictures()
        if self.request.user.is_authenticated:
            qs = Viewpoint.objects.all().distinct()
        pictures_qs = Picture.objects.order_by('-created_at')
        return qs.select_related('point').prefetch_related(
            Prefetch('pictures', queryset=pictures_qs, to_attr='ordered_pics')
        )

    def get_serializer_class(self):
        if self.action == 'list':
            if self.request.user.is_anonymous:
                return SimpleViewpointSerializer
            return SimpleAuthenticatedViewpointSerializer
        return ViewpointSerializerWithPicture

    @action(detail=False)
    def filters(self, request, *args, **kwargs):
        filter_values = {}
        for key, field in settings.TROPP_SEARCHABLE_PROPERTIES.items():
            data = None
            transform = KeyTransform(field['json_key'], 'properties')
            queryset = (Viewpoint.objects
                        .annotate(**{key: transform})
                        .exclude(**{f"{key}__isnull": True})
                        .values_list(key, flat=True))
            if field['type'] == 'single':
                # Dedupe and sort with SQL
                data = queryset.order_by(key).distinct(key)
            elif field['type'] == 'many':
                # Dedupe and sort programmatically
                data = list(queryset)
                if data:
                    data = set(reduce(operator.concat, data))
                    data = sorted(data, key=str.lower)
            if data is not None:
                filter_values[key] = data

        filter_values['photographers'] = PhotographerLabelSerializer(
            get_user_model().objects.filter(pictures__isnull=False).distinct(),
            many=True,
        ).data

        return Response(filter_values)


class CampaignViewSet(viewsets.ModelViewSet):
    queryset = Campaign.objects.all()
    permission_classes = [permissions.DjangoModelPermissions]
    http_method_names = ['get', 'post', 'put', 'delete', 'options']
    filter_backends = (CampaignFilterBackend, DateFilterBackend, SearchFilter)
    date_search_field = 'created_at'
    search_fields = ('label', )
    pagination_class = RestPageNumberPagination

    def get_queryset(self):
        # Filter only on assigned campaigns for photographs
        user = self.request.user
        pictures_qs = Picture.objects.order_by('-created_at')
        qs = super().get_queryset().prefetch_related(
            Prefetch('viewpoints__pictures', queryset=pictures_qs,
                     to_attr='ordered_pics')
        )
        if (self.action == 'list'
                and not user.has_perm('tropp.manage_all_campaigns')):
            return qs.filter(assignee=user)
        return qs

    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

    def check_object_permissions(self, request, obj: Campaign):
        # Prevent acting on unassigned campaigns for photographs
        if (not request.user.has_perm('tropp.manage_all_campaigns')
                and obj.assignee != request.user):
            self.permission_denied(request)
        super().check_object_permissions(request, obj)

    def get_serializer_class(self):
        if self.action == 'list':
            return ListCampaignNestedSerializer
        if self.action == 'retrieve':
            if self.request.user.is_anonymous:
                return DetailCampaignNestedSerializer
            return DetailAuthenticatedCampaignNestedSerializer
        return CampaignSerializer


class PictureViewSet(viewsets.ModelViewSet):
    queryset = Picture.objects.all()
    serializer_class = PictureSerializer
    permission_classes = [permissions.DjangoModelPermissions]
    pagination_class = RestPageNumberPagination

    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)



from django import template
from django.conf import settings
from django.utils.html import mark_safe

register = template.Library()


@register.simple_tag
def front_url():
    return mark_safe(settings.FRONT_URL)


@register.simple_tag
def hostname():
    return mark_safe(settings.HOSTNAME)


@register.filter
def get_item(dictionary, key):
    return dictionary.get(key)



from io import StringIO

from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase

from terracommon.terra.models import Layer
from terracommon.terra.tests.factories import FeatureFactory, LayerFactory
from terracommon.terra.tests.utils import get_files_tests


def python_function(layers_in, layer_out, *args):
    layer_out.name = "New_name"
    layer_out.save()


def python_function_raise(*args):
    raise Exception


class LayerProcessingTestCase(TestCase):
    def test_layer_processing_make_valid(self):
        layer = LayerFactory()
        output = StringIO()

        call_command(
            'layer_processing',
            f'--layer-pk-ins={layer.id}',
            f'--make-valid',
            verbosity=1, stdout=output)
        self.assertIn('The created layer pk is', output.getvalue())
        self.assertEqual(len(Layer.objects.all()), 2)

    def test_layer_processing_make_valid_polygon(self):
        layer = LayerFactory()
        output = StringIO()

        call_command(
            'layer_processing',
            f'--layer-pk-ins={layer.id}',
            f'--make-valid',
            verbosity=1, stdout=output)
        self.assertIn('The created layer pk is', output.getvalue())
        self.assertEqual(len(Layer.objects.all()), 2)

    def test_layer_processing_make_valid_fail_multiple_pk_ins(self):
        layer_1 = LayerFactory()
        layer_2 = LayerFactory()
        with self.assertRaises(ValueError) as error:
            call_command(
                'layer_processing',
                f'--layer-name-ins={layer_1.name}',
                f'--layer-name-ins={layer_2.name}',
                f'--make-valid',
                verbosity=0)
        self.assertEqual(str(error.exception), 'Exactly one input layer required')

    def test_layer_processing_rollback(self):
        layer = LayerFactory()
        output = StringIO()

        call_command(
            'layer_processing',
            f'--layer-pk-ins={layer.id}',
            f'--make-valid',
            f'--dry-run',
            verbosity=1, stdout=output)
        self.assertIn('The created layer pk is', output.getvalue())
        self.assertEqual(len(Layer.objects.all()), 1)

    def test_layer_processing_by_name(self):
        geojson = get_files_tests('toulouse.geojson')

        call_command('import_geojson', geojson, verbosity=0)

        # Retrieve the layer
        in_layer = Layer.objects.first()

        out_layer = Layer.objects.create(name='out')
        FeatureFactory(layer=out_layer, properties="Test")

        call_command(
            'layer_processing',
            f'--layer-name-ins={in_layer.name}',
            f'--layer-name-out={out_layer.name}',
            f'--sql-centroid',
            verbosity=0)

        out_layer = Layer.objects.get(name='out')
        self.assertIn('Test', [feature.properties for feature in out_layer.features.all()])
        self.assertTrue(len(out_layer.features.all()) > 0)

    def test_layer_processing_by_pk(self):
        empty_json = get_files_tests('empty.json')
        geojson = get_files_tests('toulouse.geojson')

        call_command(
            'import_geojson',
            f'{geojson}',
            f'-ls={empty_json}',
            verbosity=0)

        # Retrieve the layer
        in_layer = Layer.objects.first()

        out_layer = LayerFactory(name='out')

        call_command(
            'layer_processing',
            f'--layer-name-ins={in_layer.name}',
            f'--layer-pk-out={out_layer.pk}',
            f'--sql-centroid',
            verbosity=0)

        out_layer = Layer.objects.get(name='out')
        self.assertTrue(out_layer.features.count() > 0)

    def test_layer_processing_fail_wrong_pk_in(self):
        with self.assertRaises(CommandError) as error:
            call_command(
                'layer_processing',
                f'--layer-pk-ins=999',
                f'--layer-name-out=out_fail',
                f'--sql-centroid',
                verbosity=0)
        self.assertIn("layer-pk-ins: 999", str(error.exception))

    def test_layer_processing_fail_wrong_name_in(self):
        with self.assertRaises(CommandError) as error:
            call_command(
                'layer_processing',
                f'--layer-name-ins=in_fail',
                f'--layer-name-out=out_fail',
                f'--sql-centroid',
                verbosity=0)
        self.assertIn("layer-name-ins: in_fail", str(error.exception))

    def test_layer_processing_fail_wrong_name_out(self):
        layer = LayerFactory()
        with self.assertRaises(CommandError) as error:
            call_command(
                'layer_processing',
                f'--layer-pk-ins={layer.pk}',
                f'--layer-name-out=out_fail',
                f'--sql-centroid',
                verbosity=0)
        self.assertIn("Layer with name out_fail doesn't exist", str(error.exception))

    def test_layer_processing_fail_wrong_pk_out(self):
        layer = LayerFactory()
        with self.assertRaises(CommandError) as error:
            call_command(
                'layer_processing',
                f'--layer-pk-ins={layer.pk}',
                f'--layer-pk-out=999',
                f'--sql-centroid',
                verbosity=0)
        self.assertIn("Layer with pk 999 doesn't exist", str(error.exception))

    def test_layer_processing_clear_output(self):
        geojson = get_files_tests('toulouse.geojson')

        call_command(
            'import_geojson',
            f'{geojson}',
            verbosity=0)

        # Retrieve the layer
        in_layer = Layer.objects.first()

        out_layer = LayerFactory(name='out')
        FeatureFactory(layer=out_layer, properties="Test")

        self.assertEqual(out_layer.features.count(), 1)
        call_command(
            'layer_processing',
            f'--layer-name-ins={in_layer.name}',
            f'--layer-pk-out={out_layer.pk}',
            f'--sql-centroid',
            f'-co',
            verbosity=0)

        out_layer = Layer.objects.get(name='out')
        self.assertTrue(out_layer.features.count() > 1)
        self.assertNotIn('Test', [feature.properties for feature in out_layer.features.all()])

    def test_layer_processing_python(self):
        layer = LayerFactory()
        call_command(
            'layer_processing',
            f'--layer-name-ins={layer.name}',
            f'--python=terracommon.terra.tests.test_commands.test_layer_processing.python_function',
            verbosity=0)
        self.assertTrue(Layer.objects.filter(name="New_name").exists())

    def test_layer_processing_python_raise(self):
        layer = LayerFactory()
        with self.assertRaises(Exception):
            call_command(
                'layer_processing',
                f'--layer-name-ins={layer.name}',
                f'--python=terracommon.terra.tests.test_commands.test_layer_processing.python_function_raise',
                verbosity=0)

    def test_layer_processing_sql_like_simple_sql(self):
        layer = LayerFactory()
        call_command(
            'layer_processing',
            f'--layer-name-ins={layer.name}',
            f'--sql=SELECT identifier, properties, ST_MakeValid(geom::geometry) AS geom FROM in0',
            verbosity=0)
        self.assertEqual(len(Layer.objects.all()), 2)



from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.gis.db import models
from django.contrib.postgres.fields import JSONField
from django.utils.translation import ugettext_lazy as _
from versatileimagefield.fields import VersatileImageField

from terracommon.core.mixins import BaseUpdatableModel
from terracommon.core.settings import STATES
from terracommon.terra.models import Feature


class BaseLabelModel(BaseUpdatableModel):
    label = models.CharField(_('Label'), max_length=100)

    class Meta:
        abstract = True

    def __str__(self):
        return self.label


class ViewpointsManager(models.Manager):
    def with_accepted_pictures(self):
        return super().get_queryset().filter(
            pictures__state=settings.STATES.ACCEPTED,
        ).distinct()


class Viewpoint(BaseLabelModel):
    point = models.ForeignKey(
        Feature,
        on_delete=models.CASCADE,
        related_name='points',
    )
    properties = JSONField(_('Properties'), default=dict, blank=True)
    related = GenericRelation('datastore.RelatedDocument')

    objects = ViewpointsManager()

    @property
    def status(self):
        """
        Return the status of this viewpoint for a campaign
        :param self:
        :return: string (missing, draft, submitted, accepted)
        """
        # Get only pictures created for the campaign
        picture = self.pictures.latest()
        if picture.created_at < self.created_at:
            return settings.STATES.CHOICES_DICT[settings.STATES.MISSING]
        return STATES.CHOICES_DICT[picture.state]

    class Meta:
        permissions = (
            ('can_download_pdf', 'Is able to download a pdf document'),
        )
        ordering = ['-created_at']


class Campaign(BaseLabelModel):
    viewpoints = models.ManyToManyField(
        Viewpoint,
        related_name='campaigns',
    )
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name=_('Owner'),
        related_name='campaigns',
    )
    assignee = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name=_('Assigned to'),
        related_name='assigned_campaigns',
    )

    @property
    def statistics(self):
        queryset = self.viewpoints.annotate(
            missing=models.Count('pk', filter=models.Q(
                pictures__isnull=True
            )),
            pending=models.Count('pictures', filter=models.Q(
                pictures__state=STATES.DRAFT
            )),
            refused=models.Count('pictures', filter=models.Q(
                pictures__state=STATES.REFUSED
            )),
        ).values('missing', 'pending', 'refused')
        try:
            return queryset[0]
        except IndexError:
            return {'missing': 0, 'pending': 0, 'refused': 0}

    @property
    def status(self):
        return not self.viewpoints.exclude(
            pictures__state=STATES.ACCEPTED,
        ).exists()

    class Meta:
        permissions = (
            ('manage_all_campaigns', "Can manage all campaigns"),
        )
        ordering = ['-created_at']


class Picture(BaseUpdatableModel):
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name=_('Owner'),
        related_name='pictures',
    )
    viewpoint = models.ForeignKey(
        Viewpoint,
        on_delete=models.CASCADE,
        related_name='pictures',
    )
    # States may be : draft, metadata_ok (submitted), accepted, refused
    state = models.IntegerField(_('State'), default=STATES.DRAFT)

    properties = JSONField(_('Properties'), default=dict, blank=True)
    file = VersatileImageField(_('File'))

    # Different from created_at which is the upload date
    date = models.DateTimeField(_('Date'))

    # TODO maybe move that to another model with GenericFK?
    remarks = models.TextField(_('Remarks'), max_length=350)

    class Meta:
        permissions = (
            ('change_state_picture', 'Is able to change the picture '
                                     'state'),
        )
        # It's our main way of sorting pictures, so it better be indexed
        indexes = [
            models.Index(fields=['viewpoint', 'date']),
        ]
        get_latest_by = 'date'
        ordering = ['-date']

    def save(self, *args, **kwargs):
        if not settings.TROPP_PICTURES_STATES_WORKFLOW:
            self.state = settings.STATES.ACCEPTED
        super().save(*args, **kwargs)



from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


class TroppConfig(AppConfig):
    name = 'terracommon.tropp'

    def ready(self):
        import terracommon.tropp.signals  # noqa
        if 'versatileimagefield' not in settings.INSTALLED_APPS:
            raise ImproperlyConfigured(
                f"'{self.name}' needs 'versatileimagefield' in INSTALLED_APPS"
            )



import argparse
import json
import logging
import sys

from django.core.management import BaseCommand
from django.utils.translation import ugettext as _

from terracommon.document_generator.helpers import DocumentGenerator

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = _('Completion and document file conversion')

    possible_types = ['odt', 'pdf']

    def add_arguments(self, parser):
        parser.add_argument('--template',
                            dest='template',
                            type=argparse.FileType('rb'),
                            default=sys.stdin,
                            required=True,
                            help=_('ODT model to be completed'),
                            )
        parser.add_argument('--data',
                            dest='data',
                            help=_('Data to use to complete'),
                            )
        parser.add_argument('--type',
                            choices=self.possible_types,
                            required=True,
                            action='store',
                            dest='output_type',
                            help=_('Extension of the output file')
                            )
        parser.add_argument('--ouput',
                            dest='output_path',
                            type=argparse.FileType('wb'),
                            default=sys.stdin,
                            required=False,
                            help=_(''),
                            )

    def handle(self, *args, **options):
        dg = DocumentGenerator(options.get('template'))

        data = {}
        if options.get('data'):
            data = json.loads(options.get('data'))
        else:
            logger.warning('No data passed')

        if options.get('output_type') == 'odt':
            result = dg.get_odt(data=data)
        else:  # pdf
            result = dg.get_pdf(data=data)

        if options.get('output_path'):
            return options.get('output_path').write(result)
        else:
            return result




from rest_framework import routers

from .views import NotificationViewSet

app_name = 'notifications'

router = routers.SimpleRouter()

router.register(r'notifications',
                NotificationViewSet,
                base_name='notifications')

urlpatterns = router.urls



# Generated by Django 2.0.9 on 2018-11-20 10:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('events', '0002_eventhandler_priority'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='eventhandler',
            options={'ordering': ['id']},
        ),
    ]



import json
import logging
import uuid

from django.contrib.contenttypes.models import ContentType
from django.core.files.uploadedfile import SimpleUploadedFile
from django.db import transaction
from django.urls import reverse
from rest_framework import serializers

from terracommon.accounts.mixins import UserTokenGeneratorMixin
from terracommon.accounts.serializers import TerraUserSerializer
from terracommon.core.mixins import SerializerCurrentUserMixin
from terracommon.datastore.models import RelatedDocument
from terracommon.datastore.serializers import (RelatedDocumentPDFSerializer,
                                               RelatedDocumentSerializer)
from terracommon.document_generator.serializers import \
    DownloadableDocumentSerializer
from terracommon.events.signals import event
from terracommon.terra.models import Layer
from terracommon.terra.serializers import GeoJSONLayerSerializer

from .models import Comment, UserRequest

logger = logging.getLogger(__name__)


class UserRequestSerializer(serializers.ModelSerializer, SerializerCurrentUserMixin):
    owner = TerraUserSerializer(read_only=True)
    geojson = GeoJSONLayerSerializer(source='layer')
    reviewers = TerraUserSerializer(read_only=True, many=True)
    has_new_comments = serializers.SerializerMethodField()
    has_new_changes = serializers.SerializerMethodField()
    downloadables = DownloadableDocumentSerializer(read_only=True,
                                                   many=True,
                                                   source='downloadable')
    documents = RelatedDocumentSerializer(many=True, required=False)

    def create(self, validated_data):
        with transaction.atomic():

            layer = json.dumps(validated_data.pop('layer', {}))
            validated_data.update({
                'layer': self._create_layer(layer),
            })
            documents = validated_data.pop('documents', [])
            instance = super().create(validated_data)
            self._update_or_create_documents(instance, documents)

            try:
                instance.user_read(self.current_user)
            except AttributeError:
                logger.info('Cannot set object read since current_user is '
                            'unknown')
            return instance

    def _create_layer(self, layer_json):
        layer = Layer.objects.create(
            name=uuid.uuid4(),
            schema={},
        )

        layer.from_geojson(
            layer_json,
        )

        return layer

    def update(self, instance, validated_data):
        old_properties, old_state = instance.properties, instance.state

        if 'layer' in validated_data:
            geojson = validated_data.pop('layer')
            instance.layer.from_geojson(json.dumps(geojson), update=True)

        documents = validated_data.pop('documents', [])
        instance = super().update(instance, validated_data)
        self._update_or_create_documents(instance, documents)

        if ('state' in validated_data
                and old_state != validated_data['state']):
            event.send(
                self.__class__,
                action="USERREQUEST_STATE_CHANGED",
                user=self.context['request'].user,
                instance=instance,
                old_state=old_state)

        if ('properties' in validated_data
                and old_properties != validated_data['properties']):
            event.send(sender=self.__class__,
                       action='USERREQUEST_PROPERTIES_CHANGED',
                       user=self.context['request'].user,
                       instance=instance,
                       old_properties=old_properties)

        try:
            instance.user_read(self.current_user)
        except AttributeError:
            logger.info('Cannot set object read since current_user is '
                        'unknown')
        return instance

    def get_has_new_comments(self, obj):
        read = obj.get_user_read(self.current_user)
        last_comment = obj.get_comments_for_user(
            self.current_user).order_by('-updated_at').first()

        if read is None and last_comment is not None:
            return True

        return (last_comment is not None
                and (read.last_read < last_comment.updated_at))

    def get_has_new_changes(self, obj):
        read = obj.get_user_read(self.current_user)

        return read is None or (read.last_read < obj.updated_at)

    def _update_or_create_documents(self, instance, documents):
        for document in documents:
            RelatedDocument.objects.update_or_create(
                key=document['key'],
                object_id=instance.pk,
                content_type=ContentType.objects.get_for_model(
                                                instance.__class__),
                defaults={
                    'document': SimpleUploadedFile(document['key'],
                                                   document['document']),
                }
            )

    class Meta:
        model = UserRequest
        exclude = ('layer',)
        read_only_fields = ('owner', 'expiry', )


class CommentSerializer(serializers.ModelSerializer,
                        UserTokenGeneratorMixin):
    owner = TerraUserSerializer(read_only=True)
    attachment_url = serializers.SerializerMethodField()
    geojson = GeoJSONLayerSerializer(source='layer', required=False)

    def get_attachment_url(self, obj):
        uidb64, token = self.get_uidb64_token_for_user(self.current_user)

        if not obj.attachment:
            return None

        return "{}?uidb64={}&token={}".format(
            reverse('trrequests:comment-attachment',
                    args=[obj.userrequest_id, obj.pk]),
            uidb64,
            token)

    def create(self, validated_data):

        with transaction.atomic():
            if 'layer' in validated_data:
                layer = Layer.objects.create(
                    name=uuid.uuid4(),
                    schema={},
                )

                layer.from_geojson(
                    json.dumps(validated_data.pop('layer'))
                )

                validated_data.update({
                    'layer': layer,
                })

            instance = super().create(validated_data)
            try:
                instance.userrequest.user_read(self.context['request'].user)
            except AttributeError:
                logger.info('Cannot set object read since current_user is '
                            'unknown')
            return instance

    class Meta:
        model = Comment
        fields = '__all__'
        read_only_fields = ('owner', 'userrequest')
        extra_kwargs = {
            'attachment': {'write_only': True}
        }


class UserRequestPDFSerializer(UserRequestSerializer):
    documents = RelatedDocumentPDFSerializer(many=True, required=False)



# Generated by Django 2.0.8 on 2018-08-17 08:01

import uuid

import django.contrib.gis.db.models.fields
import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import terracommon.terra.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Feature',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('geom', django.contrib.gis.db.models.fields.GeometryField()),
                ('identifier', models.CharField(default=uuid.uuid4, max_length=255)),
                ('properties', django.contrib.postgres.fields.jsonb.JSONField()),
                ('from_date', terracommon.terra.fields.DateFieldYearLess(default='01-01', help_text='Layer validity period start')),
                ('to_date', terracommon.terra.fields.DateFieldYearLess(default='12-31', help_text='Layer validity period end')),
            ],
        ),
        migrations.CreateModel(
            name='FeatureRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('properties', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
                ('destination', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='relations_as_destination', to='terra.Feature')),
                ('origin', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='relations_as_origin', to='terra.Feature')),
            ],
        ),
        migrations.CreateModel(
            name='Layer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256)),
                ('group', models.CharField(default='__nogroup__', max_length=255)),
                ('schema', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
            ],
        ),
        migrations.CreateModel(
            name='LayerRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('schema', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
                ('destination', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='relations_as_destination', to='terra.Layer')),
                ('origin', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='relations_as_origin', to='terra.Layer')),
            ],
        ),
        migrations.AddField(
            model_name='feature',
            name='layer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='features', to='terra.Layer'),
        ),
    ]



from django.contrib.auth import get_user_model
from django.contrib.auth.forms import PasswordResetForm

UserModel = get_user_model()


class PasswordSetAndResetForm(PasswordResetForm):
    def get_users(self, email):
        """We override django get_users method since we need also users with
           unusable password"""
        active_users = UserModel._default_manager.filter(**{
            '%s__iexact' % UserModel.get_email_field_name(): email,
        })
        return (u for u in active_users)



import subprocess
from io import StringIO
from unittest import mock

from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase

from terracommon.terra.models import Feature
from terracommon.terra.tests.factories import LayerFactory
from terracommon.terra.tests.utils import get_files_tests


class ImportGeojsonTest(TestCase):

    def get_good_data(self):
        overpass_path = get_files_tests('overpass.osm')
        with open(overpass_path, 'rb') as overpass_file:
            overpass_data = overpass_file.read()
        return overpass_data

    @mock.patch('requests.get')
    def test_bad_query(self, mocked_get):
        type_feature = 'points'
        mocked_get.return_value.status_code = 400
        mocked_get.return_value.content = b"test"

        query = "bad query"
        self.assertRaises(CommandError, call_command,
                          'import_osm', f'{query}', f'-t{type_feature}')

    @mock.patch('requests.get')
    def test_overpass_down(self, mocked_get):
        type_feature = 'points'
        mocked_get.return_value.status_code = 404
        mocked_get.return_value.content = b""

        query = "good query"
        self.assertRaises(CommandError, call_command,
                          'import_osm', f'{query}', f'-t{type_feature}')

    @mock.patch('requests.get')
    def test_good_query(self, mocked_get):
        type_feature = 'points'
        mocked_get.return_value.status_code = 200
        mocked_get.return_value.content = self.get_good_data()
        query = 'good query'
        output = StringIO()
        call_command(
            'import_osm',
            f'{query}',
            f'-t{type_feature}',
            '-v 1', stderr=output)
        self.assertIn("Warning 1", output.getvalue())
        self.assertEqual(Feature.objects.count(), 2)

    @mock.patch('requests.get')
    def test_good_query_on_existing_layer(self, mocked_get):
        layer = LayerFactory()
        type_feature = 'points'
        mocked_get.return_value.status_code = 200
        mocked_get.return_value.content = self.get_good_data()
        query = 'good query'
        output = StringIO()
        call_command(
            'import_osm',
            f'{query}',
            f'-pk={layer.pk}',
            f'-t{type_feature}',
            '-v 1', stderr=output)
        self.assertIn("Warning 1", output.getvalue())
        self.assertEqual(layer.features.count(), 2)

    @mock.patch('requests.get')
    @mock.patch('terracommon.terra.management.commands.import_osm.Command.launch_cmd_ogr2ogr')
    def test_ogr2ogr_fail(self, mock_ogr2ogr_stdout, mocked_get):
        def command_ogr2ogr_fail(content, type_features):
            return '', 'Error'
        mock_ogr2ogr_stdout.side_effect = command_ogr2ogr_fail
        type_feature = 'points'
        mocked_get.return_value.status_code = 200
        mocked_get.return_value.content = self.get_good_data()
        query = 'good query'
        with self.assertRaises(CommandError) as error:
            call_command(
                'import_osm',
                f'{query}',
                f'-t{type_feature}',
                '-v 0')
        self.assertEqual("Ogr2ogr failed to create the geojson", str(error.exception))

    @mock.patch('requests.get')
    @mock.patch('subprocess.run')
    def test_ogr2ogr_fail_calledprocess(self, mock_ogr2ogr_raise, mocked_get):
        def command_ogr2ogr_fail(args, stdout, stderr, encoding):
            raise subprocess.CalledProcessError(1, 2)
        mock_ogr2ogr_raise.side_effect = command_ogr2ogr_fail
        type_feature = 'points'
        mocked_get.return_value.status_code = 200
        mocked_get.return_value.content = self.get_good_data()
        query = 'good query'
        with self.assertRaises(CommandError) as error:
            call_command(
                'import_osm',
                f'{query}',
                f'-t{type_feature}',
                '-v 0')
        self.assertEqual("Command ogr2ogr failed", str(error.exception))



import json
from io import BytesIO
from zipfile import ZipFile

from django.contrib.auth.models import Permission
from django.contrib.gis.geos import GEOSGeometry
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase
from django.urls import reverse
from rest_framework.status import (HTTP_200_OK, HTTP_201_CREATED,
                                   HTTP_204_NO_CONTENT, HTTP_400_BAD_REQUEST,
                                   HTTP_403_FORBIDDEN)
from rest_framework.test import APIClient

from terracommon.terra import GeometryTypes
from terracommon.terra.models import Feature
from terracommon.terra.tests.factories import (FeatureFactory, LayerFactory,
                                               UserFactory)
from terracommon.terra.tests.utils import get_files_tests


class LayerLineIntersectionTestCase(TestCase):
    def setUp(self):
        self.layer = LayerFactory()
        self.client = APIClient()
        self.user = UserFactory()
        self.client.force_authenticate(user=self.user)

    def test_intersect_bad_geometry(self):
        linestring = {
            "type": "LineString",
            "coordinates": [
                [
                    [1.25, 5.5, [2.65, 5.5]]
                ]
            ]
        }
        response = self.client.post(
            reverse('terra:layer-intersects', kwargs={'pk': self.layer.pk}),
            {'geom': json.dumps(linestring)},
            format='json',
        )

        self.assertEqual(HTTP_400_BAD_REQUEST, response.status_code)


class LayerPolygonIntersectTestCase(TestCase):
    def setUp(self):
        # from http://wiki.geojson.org/GeoJSON_draft_version_6#Polygon
        self.polygon = {
            "type": "Polygon",
            "coordinates": [
                [
                    [100.0, 0.0],
                    [101.0, 0.0],
                    [101.0, 1.0],
                    [100.0, 1.0],
                    [100.0, 0.0]
                ]
            ]
        }
        self.client = APIClient()
        self.user = UserFactory()
        self.client.force_authenticate(user=self.user)

    def test_polygon_intersection(self):
        layer = LayerFactory()
        Feature.objects.create(
            geom=json.dumps({
                "type": "LineString",
                "coordinates": [
                    [103.0, 0.0], [101.0, 1.0]
                ]
            }),
            layer=layer,
            properties={},
        )
        Feature.objects.create(
            geom=json.dumps({
                "type": "Point",
                "coordinates": [100.0, 0.0]
            }),
            layer=layer,
            properties={},
        )

        # This feature is not intersecting with the polygon
        # It is not expected to be in the response returned
        Feature.objects.create(
            geom=json.dumps({
                "type": "LineString",
                "coordinates": [
                    [30, 10], [10, 30], [40, 40]
                ]
            }),
            layer=layer,
            properties={},
        )

        response = self.client.post(
            reverse('terra:layer-intersects', kwargs={'pk': layer.pk}),
            {'geom': json.dumps(self.polygon)},
            format='json',
        )
        self.assertEqual(HTTP_200_OK, response.status_code)
        json_response = response.json()
        self.assertEqual(
            2,
            len(json_response['results']['features'])
        )


class LayerFeatureIntersectionTest(TestCase):
    def setUp(self):
        self.fake_geometry = {
            "type": "Point",
            "coordinates": [
                2.,
                45.
            ]
        }
        self.intersect_geometry = {
            "type": "LineString",
            "coordinates": [
                [
                    1.3839340209960938,
                    43.602521593464054
                ],
                [
                    1.4869308471679688,
                    43.60376465190968
                ]
            ]
        }
        self.intersect_ref_geometry = {
            "type": "LineString",
            "coordinates": [
                [
                    1.440925598144531,
                    43.64750394449096
                ],
                [
                    1.440582275390625,
                    43.574421623084234
                ]
            ]
        }
        self.fake_linestring = {
            "type": "LineString",
            "coordinates": [
                [
                    1.3839340209960938,
                    43.602521593464054
                ],
            ]
        }
        self.fake_polygon = {
            "type": "Polygon",
            "coordinates": [
                [
                    [
                        1.3839340209960938,
                        43.602521593464054
                    ],
                    [
                        1.440582275390625,
                        43.574421623084234
                    ]
                ]
            ]
        }

        self.group_name = 'mygroup'
        self.layer = LayerFactory.create(group=self.group_name,
                                         add_features=5)

        self.user = UserFactory()
        self.client.force_login(self.user)

    def test_features_intersections(self):
        layer = LayerFactory(group=self.group_name)
        FeatureFactory(
            layer=layer,
            geom=GEOSGeometry(json.dumps(self.intersect_ref_geometry)))

        """The layer below must intersect"""
        response = self.client.post(
            reverse('terra:layer-intersects', args=[layer.pk, ]),
            {
                'geom': json.dumps(self.intersect_geometry)
            }
        )

        self.assertEqual(HTTP_200_OK, response.status_code)
        response = response.json().get('results', {})
        self.assertEqual(
            1,
            len(response.get('features'))
        )
        self.assertDictEqual(
            self.intersect_ref_geometry,
            response.get('features')[0].get('geometry')
        )

        """The layer below must NOT intersect"""
        response = self.client.post(
            reverse('terra:layer-intersects', args=[layer.name, ]),
            {
                'geom': json.dumps(self.fake_geometry)
            }
        )

        self.assertEqual(HTTP_200_OK, response.status_code)

        response = response.json().get('results', {})
        self.assertEqual(0, len(response.get('features')))

        """Tests that the intersects view throw an error if geometry is
           invalid
        """
        response = self.client.post(
            reverse('terra:layer-intersects', args=[layer.pk, ]),
            {
                'geom': '''Invalid geometry'''
            }
        )
        self.assertEqual(HTTP_400_BAD_REQUEST, response.status_code)

    def test_features_linestring_format(self):
        response = self.client.post(
            reverse('terra:layer-intersects', args=[self.layer.pk, ]),
            {
                'geom': json.dumps(self.fake_linestring)
            }
        )

        self.assertEqual(HTTP_400_BAD_REQUEST, response.status_code)

    def test_features_polygon_format(self):
        response = self.client.post(
            reverse('terra:layer-intersects', args=[self.layer.pk, ]),
            {
                'geom': json.dumps(self.fake_polygon)
            }
        )

        self.assertEqual(HTTP_400_BAD_REQUEST, response.status_code)


class LayerShapefileTestCase(TestCase):
    def setUp(self):
        self.layer = LayerFactory()
        self.user = UserFactory()
        self.client.force_login(self.user)

    def test_shapefile_export(self):
        # Create at least one feature in the layer, so it's not empty
        self.user.user_permissions.add(Permission.objects.get(codename='can_export_layers'))
        FeatureFactory(layer=self.layer)

        shape_url = reverse('terra:layer-shapefile', args=[self.layer.pk, ])
        response = self.client.get(shape_url)
        self.assertEqual(HTTP_200_OK, response.status_code)

        zip = ZipFile(BytesIO(response.content), 'r')
        self.assertListEqual(
            sorted(['prj', 'cpg', 'shx', 'shp', 'dbf']),
            sorted([f.split('.')[1] for f in zip.namelist()])
            )

    def test_properties_serializations(self):
        layer = LayerFactory()
        test_properties = {
            'int': 42,
            'str': 'test string',
            'dict': {
                'a': 'b',
            }
        }

        serialized_properties = layer._get_serialized_properties(test_properties)
        self.assertEqual(serialized_properties['str'], test_properties['str'])
        self.assertIsInstance(serialized_properties['int'], str)
        self.assertIsInstance(serialized_properties['dict'], str)

    def test_shapefile_same_import_export(self):
        self.user.user_permissions.add(Permission.objects.get(codename='can_import_layers'))
        self.user.user_permissions.add(Permission.objects.get(codename='can_export_layers'))
        FeatureFactory(layer=self.layer, properties={
            'key1': [{
                'key3': 'hello world',
            }]
        })

        shape_url = reverse('terra:layer-shapefile', args=[self.layer.pk, ])
        response = self.client.get(shape_url)
        self.assertEqual(HTTP_200_OK, response.status_code)

        shapefile = SimpleUploadedFile('shapefile-WGS84.zip',
                                       response.content)
        new_layer = LayerFactory()
        response = self.client.post(
                reverse('terra:layer-shapefile', args=[new_layer.pk, ]),
                {'shapefile': shapefile, }
                )

        self.assertEqual(HTTP_200_OK, response.status_code)
        self.assertEqual(self.layer.features.first().properties,
                         new_layer.features.first().properties)

    def test_empty_shapefile_export(self):
        # Create en ampty layer to test its behavior
        LayerFactory()
        self.user.user_permissions.add(Permission.objects.get(codename='can_export_layers'))
        shape_url = reverse('terra:layer-shapefile', args=[self.layer.pk, ])
        response = self.client.get(shape_url)
        self.assertEqual(HTTP_204_NO_CONTENT, response.status_code)

    def test_shapefile_no_permission(self):
        shape_url = reverse('terra:layer-shapefile', args=[self.layer.pk, ])

        self.assertEqual(
            self.client.get(shape_url).status_code,
            HTTP_403_FORBIDDEN
        )

    def test_no_shapefile_import(self):
        self.user.user_permissions.add(Permission.objects.get(codename='can_import_layers'))
        layer = LayerFactory()

        response = self.client.post(
            reverse('terra:layer-shapefile', args=[layer.pk, ]),)

        self.assertEqual(HTTP_400_BAD_REQUEST, response.status_code)

    def test_shapefile_import_view(self):
        self.user.user_permissions.add(Permission.objects.get(codename='can_import_layers'))
        layer = LayerFactory()

        shapefile_path = get_files_tests('shapefile-WGS84.zip')

        with open(shapefile_path, 'rb') as fd:
            shapefile = SimpleUploadedFile('shapefile-WGS84.zip',
                                           fd.read())

            response = self.client.post(
                reverse('terra:layer-shapefile', args=[layer.pk, ]),
                {'shapefile': shapefile, }
                )

        self.assertEqual(HTTP_200_OK, response.status_code)
        self.assertEqual(8, layer.features.all().count())

    def test_shapefile_import_view_error(self):
        self.user.user_permissions.add(Permission.objects.get(codename='can_import_layers'))
        shapefile = SimpleUploadedFile('shapefile-WGS84.zip',
                                       b'bad bad data')

        response = self.client.post(
            reverse('terra:layer-shapefile', args=[self.layer.pk, ]),
            {'shapefile': shapefile, }
            )
        self.assertEqual(HTTP_400_BAD_REQUEST, response.status_code)


class LayerGeojsonTestCase(TestCase):
    def setUp(self):
        self.layer = LayerFactory()
        self.user = UserFactory()
        self.client.force_login(self.user)

    def test_to_geojson(self):
        # Create at least one feature in the layer, so it's not empty
        self.user.user_permissions.add(Permission.objects.get(codename='can_export_layers'))
        FeatureFactory(layer=self.layer)

        geojson_url = reverse('terra:layer-geojson', args=[self.layer.pk, ])
        response = self.client.get(geojson_url)

        self.assertEqual(HTTP_200_OK, response.status_code)

        response = response.json()
        self.assertEqual('FeatureCollection', response.get('type'))
        self.assertEqual(self.layer.features.all().count(),
                         len(response.get('features')))

    def test_to_geojson_no_permission(self):
        # Create at least one feature in the layer, so it's not empty
        FeatureFactory(layer=self.layer)

        geojson_url = reverse('terra:layer-geojson', args=[self.layer.pk, ])
        response = self.client.get(geojson_url)

        self.assertEqual(HTTP_403_FORBIDDEN, response.status_code)


class LayerDetailTest(TestCase):
    geometry = {
        "type": "LineString",
        "coordinates": [
            [
                1.3839340209960938,
                43.602521593464054
            ], [
                1.4869308471679688,
                43.60376465190968
            ]
        ]
    }

    def setUp(self):
        self.client = APIClient()
        self.layer = LayerFactory()
        self.user = UserFactory()
        self.client.force_authenticate(self.user)

    def test_no_permission(self):

        FeatureFactory(layer=self.layer, properties={'a': 'b'})

        response = self.client.patch(
            reverse('terra:layer-detail', args=[self.layer.name, ]), {})

        self.assertEqual(HTTP_403_FORBIDDEN, response.status_code)

    def test_update(self):
        self.user.user_permissions.add(
            Permission.objects.get(codename='can_update_features_properties')
        )
        geom = GEOSGeometry(json.dumps(self.geometry))
        feature = FeatureFactory(layer=self.layer,
                                 geom=geom,
                                 properties={'a': 'b'})

        updated_properties = {
            'c': 'd',
            'a': 'd',
            }

        response = self.client.patch(
            reverse('terra:layer-detail', args=[self.layer.name, ]),
            {
                "type": "FeatureCollection",
                "features": [
                    {
                        "type": "Feature",
                        'geometry': self.geometry,
                        'properties': updated_properties,
                    },
                ]
            }, format='json')

        self.assertEqual(HTTP_200_OK, response.status_code)
        response = response.json()
        self.assertEqual(
            response['features'][0]['properties'],
            updated_properties
            )

        feature.refresh_from_db()
        self.assertDictEqual(feature.properties, updated_properties)


class LayerCreationTest(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserFactory()
        self.client.force_authenticate(user=self.user)

        self.point_layer = LayerFactory(name="no schema point geom",
                                        geom_type=GeometryTypes.Point)
        self.null_layer = LayerFactory(name="no schema null geom", geom_type=None)

    def test_point_layer_allow_point(self):
        response = self.client.post(reverse('terra:feature-list', args=[self.point_layer.pk, ]),
                                    data={"geom": "POINT(0 0)",
                                          "properties": {"toto": "ok"}})
        self.assertEqual(response.status_code, HTTP_201_CREATED, response.json())

    def test_point_layer_disallow_other(self):
        response = self.client.post(reverse('terra:feature-list', args=[self.point_layer.pk, ]),
                                    data={"geom": "LINESTRING(0 0, 1 1)",
                                          "properties": {"toto": "ok"}})
        self.assertEqual(response.status_code, HTTP_400_BAD_REQUEST, response.json())

    def test_null_layer_allow_points(self):
        response = self.client.post(reverse('terra:feature-list', args=[self.null_layer.pk, ]),
                                    data={"geom": "POINT(0 0)",
                                          "properties": {"toto": "ok"}})
        self.assertEqual(response.status_code, HTTP_201_CREATED, response.json())

    def test_null_layer_allow_linestring(self):
        response = self.client.post(reverse('terra:feature-list', args=[self.null_layer.pk, ]),
                                    data={"geom": "LINESTRING(0 0, 1 1)",
                                          "properties": {"toto": "ok"}})
        self.assertEqual(response.status_code, HTTP_201_CREATED, response.json())



# Generated by Django 2.0.13 on 2019-03-29 10:56

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('terra', '0018_auto_20190326_1049'),
        ('terra', '0018_auto_20190228_1655'),
    ]

    operations = [
    ]



import importlib
import uuid

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractBaseUser, PermissionsMixin
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.gis.db import models
from django.contrib.postgres.fields import JSONField
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from .managers import ReadModelManager, TerraUserManager


class TerraUser(AbstractBaseUser, PermissionsMixin):
    USERNAME_FIELD = 'email'
    EMAIL_FIELD = 'email'
    REQUIRED_FIELDS = []

    uuid = models.CharField(_('unique identifier'),
                            max_length=255,
                            default=uuid.uuid4,
                            unique=True)
    email = models.EmailField(_('email address'), blank=True, unique=True)
    properties = JSONField(default=dict, blank=True)
    is_staff = models.BooleanField(
        _('staff status'),
        default=False,
        help_text=_('Designates whether the user '
                    'can log into this admin site.'),
    )
    is_active = models.BooleanField(
        _('active'),
        default=True,
        help_text=_(
            'Designates whether this user should be treated as active. '
            'Unselect this instead of deleting accounts.'
        ),
    )
    date_joined = models.DateTimeField(_('date joined'), default=timezone.now)

    objects = TerraUserManager()

    def get_by_natural_key(self, username):
        return self.get(**{f'{self.model.USERNAME_FIELD}__iexact': username})

    def __str__(self):
        try:
            path = settings.TERRA_USER_STRING_FORMAT
            module_path, attr_name = path.rsplit('.', 1)
            module = importlib.import_module(module_path)
            return getattr(module, attr_name)(self)
        except AttributeError:
            return self.email

    class Meta:
        ordering = ['id']
        permissions = (
            ('can_manage_users', 'Is able create, delete, update users'),
        )


UserModel = get_user_model()


class ReadModel(models.Model):
    user = models.ForeignKey(UserModel, on_delete=models.PROTECT)
    contenttype = models.ForeignKey(ContentType, on_delete=models.PROTECT)
    identifier = models.PositiveIntegerField()
    model_object = GenericForeignKey('contenttype', 'identifier')
    last_read = models.DateTimeField(auto_now=True)

    objects = ReadModelManager()

    def read_instance(self):
        self.save()

    class Meta:
        ordering = ['id']



from io import StringIO

from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase

from terracommon.terra.models import Layer
from terracommon.terra.tests.factories import LayerFactory
from terracommon.terra.tests.utils import get_files_tests


class UpdateTopologyTestCase(TestCase):
    def test_update_topology_routing_fail(self):
        layer = LayerFactory()
        with self.assertRaises(CommandError) as error:
            call_command(
                'update_topology',
                f'--layer-pk={layer.pk}',
                verbosity=0)
        self.assertEqual('An error occuring during topology update', str(error.exception))

    def test_update_topology_fail(self):
        output = StringIO()
        with self.assertRaises(CommandError) as error:
            call_command(
                'update_topology',
                f'--layer-pk=999',
                verbosity=0, stdout=output)
        self.assertEqual("Layer with pk 999 doesn't exist", str(error.exception))

    def test_update_topology(self):
        geojson = get_files_tests('toulouse.geojson')

        call_command(
            'import_geojson',
            f'{geojson}',
            verbosity=0)

        # Retrieve the layer
        in_layer = Layer.objects.first()

        output = StringIO()
        call_command(
            'update_topology',
            f'--layer-pk={in_layer.pk}',
            verbosity=1, stdout=output)
        self.assertIn('Topology successfully updated', output.getvalue())

    def test_update_topology_rollback(self):
        geojson = get_files_tests('toulouse.geojson')

        call_command(
            'import_geojson',
            f'{geojson}',
            verbosity=0)

        # Retrieve the layer
        in_layer = Layer.objects.first()

        output = StringIO()
        call_command(
            'update_topology',
            '--dry-run',
            f'--layer-pk={in_layer.pk}',
            verbosity=1, stdout=output)



# Generated by Django 2.0.13 on 2019-05-03 15:26

from django.db import migrations, models
import terracommon.terra.models


class Migration(migrations.Migration):

    dependencies = [
        ('terra', '0022_auto_20190417_1610'),
    ]

    operations = [
        migrations.AddField(
            model_name='layer',
            name='geom_type',
            field=models.IntegerField(choices=[(terracommon.terra.models.GeometryTypes(0), 0), (terracommon.terra.models.GeometryTypes(1), 1), (terracommon.terra.models.GeometryTypes(3), 3), (terracommon.terra.models.GeometryTypes(4), 4), (terracommon.terra.models.GeometryTypes(5), 5), (terracommon.terra.models.GeometryTypes(6), 6), (terracommon.terra.models.GeometryTypes(7), 7)], null=True),
        ),
    ]



from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase

from terracommon.terra.models import Layer
from terracommon.terra.tests.utils import get_files_tests


class LayerEditTest(TestCase):
    def setUp(self):
        shapefile_path = get_files_tests('shapefile-WGS84.zip')
        sample_shapefile = open(shapefile_path, 'rb')
        call_command(
            'import_shapefile', sample_shapefile.name,
            '-i', 'ID_PG',
            verbosity=0)

    def test_layer_edit(self):
        # Ensure old settings
        layer = Layer.objects.all()[0]
        self.assertNotEqual('new_name', layer.name)
        self.assertNotEqual('new_group', layer.group)
        self.assertNotEqual({'foo': 'bar'}, layer.settings)

        foo_bar_json = get_files_tests('foo_bar.json')

        # Change settings
        call_command(
            'layer_edit',
            '-pk', layer.pk,
            '-l', 'new_name',
            '-gr', 'new_group',
            '-ls', foo_bar_json
        )

        # Ensure new settings
        layer = Layer.objects.all()[0]
        self.assertEqual('new_name', layer.name)
        self.assertEqual('new_group', layer.group)
        self.assertEqual({'foo': 'bar'}, layer.settings)

    def test_layer_edit_fail_wrong_pk(self):
        # Ensure old settings

        foo_bar_json = get_files_tests('foo_bar.json')
        # Change settings
        with self.assertRaises(CommandError):
            call_command(
                'layer_edit',
                '-pk', 999,
                '-l', 'new_name',
                '-gr', 'new_group',
                '-s', foo_bar_json,
                '-ls', foo_bar_json
            )

    def test_layer_edit_with_bad_schema(self):
        layer = Layer.objects.first()
        bad_json = get_files_tests('bad.json')
        # Change settings
        with self.assertRaises(CommandError):
            call_command(
                'layer_edit',
                '-pk', layer.pk,
                '-l', 'new_name',
                '-gr', 'new_group',
                '-s', bad_json,
                '-ls', bad_json
            )

    def test_layer_edit_schema_bad_file(self):
        layer = Layer.objects.first()
        bad_file = get_files_tests('shapefile-RFG93.zip')
        # Change settings
        with self.assertRaises(CommandError):
            call_command(
                'layer_edit',
                '-pk', layer.pk,
                '-l', 'new_name',
                '-gr', 'new_group',
                '-s', bad_file,
                '-ls', bad_file
            )

    def test_layer_edit_bad_settings(self):
        layer = Layer.objects.first()
        bad_json = get_files_tests('bad.json')
        # Change settings
        with self.assertRaises(CommandError):
            call_command(
                'layer_edit',
                '-pk', layer.pk,
                '-l', 'new_name',
                '-gr', 'new_group',
                '-ls', bad_json
            )

    def test_layer_edit_guess_zoom(self):
        layer = Layer.objects.first()
        # Change settings
        call_command(
            'layer_edit',
            '-pk', layer.pk,
            '-l', 'new_name',
            '-gr', 'new_group',
            '-gz'
        )
        self.assertEqual(str(layer.settings), "{'tiles': {'maxzoom': 15, 'minzoom': 15}}")



from io import StringIO

from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase

from terracommon.terra.models import Layer
from terracommon.terra.tests.factories import LayerFactory
from terracommon.terra.tests.utils import get_files_tests


class ImportGeojsonTest(TestCase):
    def test_default_group(self):
        output = StringIO()
        call_command(
            'import_geojson', get_files_tests('empty.json'),
            verbosity=1, stdout=output)

        # Retrieve the layer
        layer = Layer.objects.first()
        self.assertIn(f'The created layer pk is {layer.pk}', output.getvalue())
        self.assertEqual('__nogroup__', layer.group)

    def test_default_group_nogroup_rollback(self):
        output = StringIO()
        call_command(
            'import_geojson',
            f'--dry-run', get_files_tests('empty.json'),
            verbosity=1, stdout=output)
        self.assertIn("The created layer pk is", output.getvalue())
        # Retrieve the layer
        self.assertEqual(Layer.objects.count(), 0)

    def test_schema_generated(self):
        call_command(
            'import_geojson',
            get_files_tests('bati.geojson'),
            '-gs',
            verbosity=0)

        # Retrieve the layer
        layer = Layer.objects.get()

        # Assert schema properties are presents
        self.assertNotEqual(
            layer.schema.get('properties').keys() -
            ['ALTITUDE', 'ETIQUETTE', 'HAUTEUR', 'ID', 'ID_PG', 'NATURE', 'NOM',
             'ORIGIN_BAT', 'PUB_XDECAL', 'PUB_YDECAL', 'ROTATION', 'ROTATION_S',
             'XDECAL', 'XDECAL_SYM', 'YDECAL', 'YDECAL_SYM', 'Z_MAX', 'Z_MIN', ], True)

    def test_import_geojson_layer_with_bad_settings(self):
        empty_geojson = get_files_tests('empty.json')
        bad_json = get_files_tests('bad.json')
        with self.assertRaises(CommandError) as error:
            call_command(
                'import_geojson',
                empty_geojson,
                '-ls', bad_json,
                verbosity=0)
        self.assertEqual("Please provide a valid layer settings file", str(error.exception))

    def test_import_geojson_layer_with_pk_layer(self):
        layer = LayerFactory()
        self.assertEqual(len(layer.features.all()), 0)
        geojson_sample = get_files_tests('toulouse.geojson')
        call_command(
            'import_geojson',
            f'--layer-pk={layer.pk}',
            geojson_sample,
            verbosity=0
        )
        self.assertEqual(len(layer.features.all()), 838)

    def test_import_geojson_layer_with_wrong_pk_layer(self):
        geojson_sample = get_files_tests('toulouse.geojson')
        with self.assertRaises(CommandError) as error:
            call_command(
                'import_geojson',
                f'--layer-pk=999',
                '-gs', geojson_sample,
                verbosity=0
            )
        self.assertIn("Layer with pk 999 doesn't exist", str(error.exception))



from django.urls import reverse
from rest_framework import serializers

from terracommon.accounts.mixins import UserTokenGeneratorMixin

from .models import DocumentTemplate, DownloadableDocument


class DownloadableDocumentSerializer(serializers.ModelSerializer,
                                     UserTokenGeneratorMixin):
    title = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()

    class Meta:
        model = DownloadableDocument
        fields = ('title', 'url', )

    def get_title(self, obj):
        return obj.document.name

    def get_url(self, obj):
        uidb64, token = self.get_uidb64_token_for_user(self.current_user)
        return "{}?uidb64={}&token={}".format(
            reverse('document_generator:document-pdf', kwargs={
                'request_pk': obj.linked_object.id,
                'pk': obj.document.id

            }),
            uidb64,
            token,
        )


class DocumentTemplateSerializer(serializers.ModelSerializer,
                                 UserTokenGeneratorMixin):
    url = serializers.SerializerMethodField()

    def get_url(self, obj):

        uidb64, token = self.get_uidb64_token_for_user(self.current_user)
        return "{}?uidb64={}&token={}".format(
            reverse('document_generator:document-file', kwargs={
                'pk': obj.id

            }),
            uidb64,
            token,
        )

    class Meta:
        model = DocumentTemplate
        fields = '__all__'
        extra_kwargs = {
            'documenttemplate': {'write_only': True}
        }



from rest_framework_jwt.utils import jwt_payload_handler

from .serializers import TerraUserSerializer


def terra_payload_handler(user):
    """ Custom response payload handler.

    This function controlls the custom payload after login or token refresh.
    This data is returned through the web API.
    """
    payload = jwt_payload_handler(user)

    user_serializer = TerraUserSerializer(user)
    payload.update({
        'user': user_serializer.data
    })
    return payload



from django.conf import settings
from django.db.models.signals import post_save
from django.dispatch import receiver
from versatileimagefield.serializers import VersatileImageFieldSerializer

from .models import Viewpoint


@receiver(post_save, sender=Viewpoint)
def update_or_create_viewpoint(instance, **kwargs):
    """
    This signal is triggered after Viewpoint save (update or create). It will
    update its feature with the related viewpoint details.
    """
    point = instance.point
    point.properties = {
        'viewpoint_id': instance.id,
        'viewpoint_label': instance.label,
    }

    # Add thumbnail representation in the feature's properties
    if instance.pictures.exists():
        last_picture_sizes = VersatileImageFieldSerializer(
            'tropp'
        ).to_representation(instance.pictures.latest().file)
        point.properties['viewpoint_picture'] = last_picture_sizes['thumbnail']

    # Add any specified viewpoint property in the feature's properties
    for prop in settings.TROPP_FEATURES_PROPERTIES_FROM_VIEWPOINT:
        value = instance.properties.get(prop)
        if value is not None:
            point.properties[f'viewpoint_{prop}'] = value

    point.save()




from django.contrib.gis.serializers import geojson as basegeojson


class Serializer(basegeojson.Serializer):
    def _init_options(self):
        super()._init_options()
        self.properties_field = self.json_kwargs.pop('properties_field', None)

    def handle_field(self, obj, field):
        if field.name == self.properties_field:
            self._current = field.value_from_object(obj)
        else:
            super().handle_field(obj, field)



import mimetypes
import os
import tempfile
import zipfile
from io import StringIO
from os.path import basename
from urllib.parse import urlparse

import weasyprint
from django.conf import settings
from django.contrib.staticfiles.finders import find
from django.core.files.storage import default_storage
from rest_framework import renderers

from .helpers import CustomCsvBuilder


class CSVRenderer(renderers.BaseRenderer):
    media_type = 'text/csv'
    format = 'csv'

    def render(self, data, media_type=None, renderer_context=None):
        csvbuilder = CustomCsvBuilder(data)
        csv_file = StringIO()

        csvbuilder.create_csv(csv_file)

        return csv_file.read()


def django_url_fetcher(url):
    mime_type, encoding = mimetypes.guess_type(url)
    url_path = urlparse(url).path
    data = {
        'mime_type': mime_type,
        'encoding': encoding,
        'filename': basename(url_path),
    }

    minio_scheme = 'https' if settings.AWS_S3_SECURE_URLS else 'http'
    minio_url_prefix = f"{minio_scheme}://{settings.AWS_S3_CUSTOM_DOMAIN}"

    if url.startswith(minio_url_prefix):  # media file from minio
        url = url.replace(
            minio_url_prefix,
            settings.AWS_S3_ENDPOINT_URL + settings.AWS_STORAGE_BUCKET_NAME
        )

    elif url_path.startswith(settings.MEDIA_URL):  # media file from filesystem
        path = url_path.replace(settings.MEDIA_URL, settings.MEDIA_ROOT)
        data['file_obj'] = default_storage.open(path)
        return data

    elif url_path.startswith(settings.STATIC_URL):
        path = url_path.replace(settings.STATIC_URL, '')
        data['file_obj'] = open(find(path), 'rb')
        return data

    # fall back to weasyprint default fetcher
    return weasyprint.default_url_fetcher(url)


class PdfRenderer(renderers.TemplateHTMLRenderer):
    media_type = 'application/pdf'
    format = 'pdf'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """Returns the rendered pdf"""
        html = super().render(data, accepted_media_type=accepted_media_type,
                              renderer_context=renderer_context)
        request = renderer_context['request']
        base_url = request.build_absolute_uri("/")

        kwargs = {}
        return weasyprint.HTML(
            string=html,
            base_url=base_url,
            url_fetcher=django_url_fetcher,
        ).write_pdf(**kwargs)


class ZipRenderer(renderers.BaseRenderer):
    media_type = 'application/zip'
    format = 'zip'

    def render(self, data, media_type=None, renderer_context=None):
        with tempfile.SpooledTemporaryFile() as tmp:
            with zipfile.ZipFile(tmp, 'w', zipfile.ZIP_DEFLATED) as archive:
                for file in data:
                    archive.writestr(
                        os.path.basename(file.name),
                        file.open().read(),
                    )
            tmp.seek(0)
            return tmp.read()



from django.db.models import Q, QuerySet

from . import models as ds_models


class DataStoreQuerySet(QuerySet):
    def get_datastores_for_user(self, user, perms=None):
        prefixes = self._get_user_prefixes_query(user, perms)

        return self.filter(prefixes)

    def _get_user_prefixes_query(self, user, perms=None):
        prefixes_query = Q()
        query_args = {
            'group__in': user.groups.all(),
        }
        if perms:
            query_args['permission__in'] = perms

        permissions = ds_models.DataStorePermission.objects.filter(
                                                            **query_args)
        if not permissions:
            return Q(key=None)

        for ds_perm in permissions:
            prefixes_query |= Q(key__startswith=ds_perm.prefix)

        return prefixes_query



from django.contrib.auth import get_user_model

UserModel = get_user_model()


def users_of_group(group_name):
    return [
        {
            'email': user.email,
            'properties': user.properties,
        }
        for user in UserModel.objects.filter(groups__name=group_name)
    ]


def get_user_from_pk(pk):
    try:
        user = UserModel.objects.get(pk=pk)
        return {
            'email': user.email,
            'properties': user.properties,
        }
    except UserModel.DoesNotExist:
        return



# Generated by Django 2.0.13 on 2019-07-15 12:10

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('datastore', '0003_auto_20181120_1059'),
    ]

    operations = [
        migrations.AddField(
            model_name='relateddocument',
            name='properties',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]
